/*
 *  scst_ipc_cache_ctrl.c
 *
 *  Copyright (C) 2013 Sony Corporation
 *
 *  SCSI dev handler using IPCIF.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, version 2
 *  of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>

#include <linux/kthread.h>
#include <linux/delay.h>

#ifdef INSIDE_KERNEL_TREE
#include <scst/scst.h>
#else
#include "scst.h"
#endif

#include "scst_dev_handler.h"

#include "scst_pcimem.h"
#include "scst_ipc.h"

#if defined(CONFIG_SCST_DEBUG) || defined(CONFIG_SCST_TRACING)

static unsigned long *p_dh_trace_flag;

#undef trace_flag
#define trace_flag (*p_dh_trace_flag)

#endif /* defined(CONFIG_SCST_DEBUG) || defined(CONFIG_SCST_TRACING) */


static inline int scstipc_is_cache_cmd(const struct scst_cmd *cmd)
{
	if ((cmd->cdb[0] == 0x00) ||
		(cmd->cdb[0] == 0x12)) {
		return 1;
	}

	return 0;
}

static inline int scstipc_is_media_check_cmd(const struct scst_cmd *cmd)
{
	if (cmd->cdb[0] == 0x00) {
		return 1;
	}

	return 0;
}

static inline int get_evpd(const struct scst_cmd *cmd, uint8_t *evpd, uint8_t *page)
{
	*evpd = 0;
	*page = 0;

	switch (cmd->cdb[0]) {
	case 0x12: /* INQUIRY */
		*evpd = (cmd->cdb[1] & 0x01);
		if (*evpd) {
			*page = cmd->cdb[2];
		}
		break;
	default:
		break;
	}

	return 0;
}

static int scstipc_cache_alloc_local_buf(int len, struct scatterlist *sg, int *sg_cnt)
{
	int size = 0;
	int i;

	TRACE_ENTRY();

	for (i = 0; size < len; i++) {
		struct page *page = alloc_pages(GFP_KERNEL, 0);
		sg_set_page(&sg[i], page, PAGE_SIZE, 0);

		size += PAGE_SIZE;
	}

	*sg_cnt += i;

	TRACE_EXIT();
	return 0;
}

void scstipc_cache_free_local_buf(struct scatterlist *sg, int sg_cnt)
{
	int i;

	TRACE_ENTRY();

	for (i = 0; i < sg_cnt; i++) {
		struct page *p = sg_page(&sg[i]);
		int len = sg[i].length;
		int pages =
			(len >> PAGE_SHIFT) + ((len & ~PAGE_MASK) != 0);

		while (pages > 0) {
			__free_pages(p, 0);

			pages--;
			p++;
		}
	}

	TRACE_EXIT();

	return;
}

static void scstipc_clear_cache(struct scst_pcimem_dev *virt_dev)
{
	struct scst_cache_data *cache_data;
	int ret;

	TRACE_ENTRY();

	TRACE_MEM("clearing cache in %s", virt_dev->name);

	ret = down_interruptible(&virt_dev->cache_sem);
	if (ret) {
		goto out;
	}

	while (!list_empty(&virt_dev->cache_list)) {
		cache_data = list_entry(virt_dev->cache_list.next, struct scst_cache_data, list_entry);

		list_del(&cache_data->list_entry);

		TRACE_MEM("freeing cache_data->data=%p", cache_data->data);
		kfree(cache_data->data);

		TRACE_MEM("freeing cache_data=%p", cache_data);
		kfree(cache_data);
	}

	up(&virt_dev->cache_sem);

out:
	TRACE_EXIT();
	return;
}

static struct scst_cache_data *scstipc_search_cache(struct scst_cmd *cmd)
{
	struct scst_device *dev = cmd->dev;
	struct scst_pcimem_dev *virt_dev = dev->dh_priv;
	struct scst_cache_data *cache_data;
	struct scst_cache_data *res = NULL;
	uint8_t evpd;
	uint8_t page;

	TRACE_ENTRY();

	get_evpd(cmd, &evpd, &page);

	list_for_each_entry(cache_data, &virt_dev->cache_list, list_entry) {
		if ((cache_data->opcode == cmd->cdb[0]) &&
			(cache_data->evpd == evpd) &&
			(cache_data->page == page)) {
			res = cache_data;
		}
	}

	TRACE_EXIT();
	return res;
}

static struct scst_cache_data *scstipc_create_cache(struct scst_cmd *cmd)
{
	struct scst_cache_data *cache_data = NULL;
	uint8_t *data;
	uint8_t *virt_base;
	int i, j;

	TRACE_ENTRY();

	data = kmalloc(cmd->resp_data_len, GFP_KERNEL);
	if (!data) {
		PRINT_ERROR("kmalloc(0x%x) failed.", cmd->resp_data_len);
		goto out;
	}

	TRACE_MEM("alloced cache_data->data=%p", data);

	cache_data = kzalloc(sizeof(*cache_data), GFP_KERNEL);
	if (!cache_data) {
		PRINT_ERROR("kmalloc(0x%x) failed.", sizeof(*cache_data));
		goto out_free;
	}

	TRACE_MEM("alloced cache_data=%p", cache_data);

	cache_data->opcode = cmd->cdb[0];
	get_evpd(cmd, &cache_data->evpd, &cache_data->page);
	if (cmd->sense) {
		cache_data->resp_code = cmd->sense[0];
		cache_data->sense_key = cmd->sense[2];
		cache_data->asc = cmd->sense[12];
		cache_data->ascq = cmd->sense[13];
	}
	cache_data->data_len = cmd->resp_data_len;
	cache_data->data = data;

	j = 0;
	for (i = 0; i < cmd->sg_cnt; i++) {
		virt_base = ioremap_nocache(pcimem_bus_to_phys(cmd->sg[i].dma_address), cmd->sg[i].dma_length);
		if (!virt_base) {
			PRINT_ERROR("ioremap_nocache(0x%08x,0x%08x) failed.", cmd->sg[i].dma_address, cmd->sg[i].dma_length);
			goto out_free;
		}

		memcpy(&cache_data->data[j], virt_base, cmd->sg[i].dma_length);
		iounmap(virt_base);
		j += cmd->sg[i].dma_length;
	}

	TRACE(TRACE_DEBUG, "opcode=0x%02x,evpd=%d,page=0x%02x,data_len=%d",
		  cache_data->opcode,
		  cache_data->evpd,
		  cache_data->page,
		  cache_data->data_len);

	TRACE_BUFFER("created cache data", cache_data->data, cache_data->data_len);

out:
	TRACE_EXIT();
	return cache_data;

out_free:
	TRACE_MEM("freeing cache_data->data=%p", data);
	kfree(data);

	if (cache_data) {
		TRACE_MEM("freeing cache_data=%p", cache_data);
		kfree(cache_data);
		cache_data = NULL;
	}

	goto out;
}

static int scstipc_update_cache(struct scst_cmd *cmd)
{
	struct scst_device *dev = cmd->dev;
	struct scst_pcimem_dev *virt_dev = dev->dh_priv;
	struct scst_cache_data *old_cache_data, *new_cache_data;
	int res = 0;
	int ret;

	TRACE_ENTRY();

	new_cache_data = scstipc_create_cache(cmd);
	if (!new_cache_data) {
		goto out;
	}

	ret = down_interruptible(&virt_dev->cache_sem);
	if (ret) {
		goto out_free;
	}

	old_cache_data = scstipc_search_cache(cmd);
	if (old_cache_data) {
		list_del(&old_cache_data->list_entry);
	}

	list_add_tail(&new_cache_data->list_entry, &virt_dev->cache_list);

	up(&virt_dev->cache_sem);

	if (old_cache_data) {
		TRACE_MEM("freeing old_cache_data->data=%p", old_cache_data->data);
		kfree(old_cache_data->data);

		TRACE_MEM("freeing old_cache_data=%p", old_cache_data);
		kfree(old_cache_data);
	}

out:
	TRACE_EXIT_RES(res);
	return res;

out_free:
	TRACE_MEM("freeing cache_data->data=%p", new_cache_data->data);
	kfree(new_cache_data->data);

	TRACE_MEM("freeing cache_data=%p", new_cache_data);
	kfree(new_cache_data);
	goto out;
}

void scstipc_check_save_cache(struct scst_cmd *cmd)
{
	struct scst_device *dev = cmd->dev;
	struct scst_pcimem_dev *virt_dev = dev->dh_priv;

	TRACE_ENTRY();

	if (cmd->sense &&
		cmd->sense[2] == UNIT_ATTENTION) {
		/* ignore UNIT_ATTENTION*/
		goto out;
	}

	if (scstipc_is_media_check_cmd(cmd) &&
		cmd->sense) {
		/* Medium Not Ready */
		virt_dev->media_exists = 0;
		goto out;
	}

	switch (cmd->cdb[0]) {
	case 0x00: /* TEST UNIT READY */
		virt_dev->media_exists = 1;
		break;
	case 0x12: /* INQUIRY */
		scstipc_update_cache(cmd);
		break;
	default:
		break;
	}

out:
	TRACE_EXIT();
	return;
}

int scstipc_check_read_cache(struct scst_cmd *cmd)
{
	struct scst_device *dev = cmd->dev;
	struct scst_pcimem_dev *virt_dev = dev->dh_priv;
	struct scst_cache_data *cache_data;
	uint8_t *buf;
	int len;
	int res = 0;
	int ret;

	TRACE_ENTRY();

	if (!scstipc_is_cache_cmd(cmd)) {
		res = -1;
		goto out;
	}

	if (scstipc_is_media_check_cmd(cmd) &&
		!virt_dev->media_exists) {
		TRACE_DBG_FLAG(TRACE_BUFF, "returning CHECK_CONDITION(02/3a/00) cmd=%p, opcode=0x%02x, LUN=%llu, initiator=%s", cmd, cmd->cdb[0], cmd->lun, cmd->sess->initiator_name);
		scst_set_cmd_error(cmd, NOT_READY, 0x3a, 0x00);
		goto out;
	}

	if (cmd->cdb[0] == 0x00) { /* TEST UNIT READY */
		/* OK, good condition */
		TRACE_DBG_FLAG(TRACE_BUFF, "returning GOOD_CONDITION cmd=%p, opcode=0x%02x, LUN=%llu, initiator=%s", cmd, cmd->cdb[0], cmd->lun, cmd->sess->initiator_name);

		goto out;
	}

	ret = down_interruptible(&virt_dev->cache_sem);
	if (ret) {
		res = -1;
		goto out;
	}

	cache_data = scstipc_search_cache(cmd);
	if (!cache_data) {
		res = -1;
		goto out_unlock;
	}

	if (cache_data->resp_code) {
		scst_set_cmd_error(cmd,
						   cache_data->sense_key,
						   cache_data->asc,
						   cache_data->ascq);

		goto out_unlock;
	}

	scstipc_cache_alloc_local_buf(cache_data->data_len, cmd->sg, &cmd->sg_cnt);

	len = scst_get_buf_full(cmd, &buf);

	memcpy(buf, cache_data->data, cache_data->data_len);

	TRACE_DBG_FLAG(TRACE_BUFF, "cmd=%p, opcode=0x%02x, LUN=%llu, initiator=%s", cmd, cmd->cdb[0], cmd->lun, cmd->sess->initiator_name);

	TRACE_BUFFER("read cache-data", buf, cache_data->data_len);

	scst_put_buf_full(cmd, buf);

out_unlock:
	up(&virt_dev->cache_sem);

out:
	TRACE_EXIT_RES(res);
	return res;
}

int scstipc_init_dev_cache(struct scst_pcimem_dev *virt_dev)
{
	sema_init(&virt_dev->cache_sem, 1);

	INIT_LIST_HEAD(&virt_dev->cache_list);

	return 0;
}

void scstipc_exit_dev_cache(struct scst_pcimem_dev *virt_dev)
{
	scstipc_clear_cache(virt_dev);

	return;
}

int scstipc_init_cache_ctrl(void)
{
	p_dh_trace_flag = get_trace_flag_ptr();
	dh_trace_flag = trace_flag; // unused

	return 0;
}

void scstipc_exit_cache_ctrl(void)
{
	return;
}

/* end of scst_ipc_cache_ctrl.c */
