/*
 *  scst_ipc_delayed_work.c
 *
 *  Copyright (C) 2013 Sony Corporation
 *
 *  SCSI dev handler using IPCIF.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, version 2
 *  of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>

#include <linux/kthread.h>
#include <linux/spinlock.h>
#include <linux/delay.h>

#ifdef INSIDE_KERNEL_TREE
#include <scst/scst.h>
#else
#include "scst.h"
#endif

#include "scst_dev_handler.h"

#include "../../../../sony/ipcif_scstlib/ipcif_scstlib.h"
#include "scst_pcimem.h"
#include "scst_ipc.h"


static struct task_struct *scstipc_send_thread = NULL;
static spinlock_t scstipc_lock;
static wait_queue_head_t scstipc_waitqueue;
static unsigned int scstipc_pending_work;
#define SCSTIPC_WORK_STOP (0x01 << 0)
#define SCSTIPC_WORK_DMADONE (0x01 << 1)
#define SCSTIPC_WORK_RESET (0x01 << 2)

struct scstipc_dmadone_work {
	struct ipcscst_areainfo areainfo[MAX_SGV_CNT];
	int sg_cnt;
	int mbxid;
	unsigned int tag;
	unsigned int direction;
	struct list_head list;
};

struct scstipc_reset_work {
	int mbxid;
	struct list_head list;
};

static LIST_HEAD(scstipc_dmadone_list);
static LIST_HEAD(scstipc_reset_list);

static unsigned int scstipc_test_and_clear_pending_work(void)
{
	unsigned long irqflags;
	unsigned int work;

	spin_lock_irqsave(&scstipc_lock, irqflags);

	work = scstipc_pending_work;
	scstipc_pending_work = 0;

	spin_unlock_irqrestore(&scstipc_lock, irqflags);

	return work;
}

static void scstipc_set_pending_work(unsigned int work)
{
	unsigned long irqflags;

	spin_lock_irqsave(&scstipc_lock, irqflags);
	scstipc_pending_work |= work;
	spin_unlock_irqrestore(&scstipc_lock, irqflags);

	wake_up_interruptible(&scstipc_waitqueue);
}

int scstipc_add_dmadone_work(const struct scst_cmd *cmd)
{
	struct scstipc_dmadone_work *dmadone_work;
	int atomic = in_atomic();
	const struct scst_device *dev = cmd->dev;
	const struct scst_pcimem_dev *virt_dev = dev->dh_priv;
	unsigned long irqflags;
	int i;

	dmadone_work = kmalloc(sizeof(*dmadone_work), atomic ? GFP_ATOMIC : GFP_KERNEL);
	if (!dmadone_work) {
		PRINT_ERROR("%s", "kmalloc() error.");
		return -1;
	}

	//	PRINT_INFO("kmalloc(): %p\n", dmadone_work);

	for (i = 0; i < cmd->dh_sg_cnt; i++) {
		dmadone_work->areainfo[i].addr = cmd->dh_sg[i].dma_address;
		dmadone_work->areainfo[i].size = cmd->dh_sg[i].dma_length;
	}

	dmadone_work->mbxid = virt_dev->ibmbxid;
	dmadone_work->tag = cmd->dh_sn;
	dmadone_work->sg_cnt = cmd->dh_sg_cnt;
	switch (cmd->data_direction) {
	case SCST_DATA_READ:
		dmadone_work->direction = IPCSCST_DATA_READ;
		break;
	case SCST_DATA_WRITE:
		dmadone_work->direction = IPCSCST_DATA_WRITE;
		break;
	default:
		dmadone_work->direction = 0;
		break;
	}

	//	PRINT_INFO("opcode=0x%02x, direction=%d\n", cmd->cdb[0], dmadone_work->direction);

	spin_lock_irqsave(&scstipc_lock, irqflags);

	list_add_tail(&dmadone_work->list, &scstipc_dmadone_list);

	spin_unlock_irqrestore(&scstipc_lock, irqflags);

	scstipc_set_pending_work(SCSTIPC_WORK_DMADONE);

	return 0;
}

int scstipc_add_reset_work(struct scst_mgmt_cmd *mcmd, struct scst_tgt_dev *tgt_dev)
{
	struct scstipc_reset_work *reset_work;
	int atomic = in_atomic();
	const struct scst_device *dev = tgt_dev->dev;
	const struct scst_pcimem_dev *virt_dev = dev->dh_priv;
	unsigned long irqflags;

	reset_work = kmalloc(sizeof(*reset_work), atomic ? GFP_ATOMIC : GFP_KERNEL);
	if (!reset_work) {
		PRINT_ERROR("%s", "kmalloc() error.");
		return -1;
	}

	//	PRINT_INFO("kmalloc(): %p\n", reset_work);


	reset_work->mbxid = virt_dev->ibmbxid;

	spin_lock_irqsave(&scstipc_lock, irqflags);

	list_add_tail(&reset_work->list, &scstipc_reset_list);

	spin_unlock_irqrestore(&scstipc_lock, irqflags);

	scstipc_set_pending_work(SCSTIPC_WORK_RESET);

	return 0;
}

static int scstipc_proc_dmadone_work(void)
{
	struct scstipc_dmadone_work *e;
	unsigned long irqflags;
	int ret;

	spin_lock_irqsave(&scstipc_lock, irqflags);

	while (!list_empty(&scstipc_dmadone_list)) {
		e = list_entry(scstipc_dmadone_list.next, struct scstipc_dmadone_work, list);
		list_del(&e->list);
		spin_unlock_irqrestore(&scstipc_lock, irqflags);

		//		PRINT_INFO("ipcscst_send_dmadone(mbxid=%d)\n", e->mbxid);
		ret = ipcscst_send_dmadone(e->mbxid, e->tag, e->areainfo, e->sg_cnt, e->direction);
		if (ret < 0) {
			PRINT_ERROR("ipcscst_send_dmadone() error(%d).\n", ret);
		}

		//		PRINT_INFO("kfree(): %p\n", e);

		kfree(e);

		spin_lock_irqsave(&scstipc_lock, irqflags);
	}

	spin_unlock_irqrestore(&scstipc_lock, irqflags);

	return 0;
}

static int scstipc_proc_reset_work(void)
{
	struct scstipc_reset_work *e;
	unsigned long irqflags;
	int ret;

	spin_lock_irqsave(&scstipc_lock, irqflags);

	while (!list_empty(&scstipc_reset_list)) {
		e = list_entry(scstipc_reset_list.next, struct scstipc_reset_work, list);
		list_del(&e->list);
		spin_unlock_irqrestore(&scstipc_lock, irqflags);

		//		PRINT_INFO("ipcscst_send_reset(mbxid=%d)\n", e->mbxid);
		ret = ipcscst_send_reset(e->mbxid);
		if (ret < 0) {
			PRINT_ERROR("ipcscst_send_reset() error(%d).\n", ret);
		}

		//		PRINT_INFO("kfree(): %p\n", e);

		kfree(e);

		spin_lock_irqsave(&scstipc_lock, irqflags);
	}

	spin_unlock_irqrestore(&scstipc_lock, irqflags);

	return 0;
}

static int scstipc_do_send_thread(void *data)
{
	unsigned int pending_work;
	int ret;

	while (1) {
		ret = wait_event_interruptible(scstipc_waitqueue, (pending_work = scstipc_test_and_clear_pending_work()) != 0);

		if (pending_work & SCSTIPC_WORK_DMADONE) {
			scstipc_proc_dmadone_work();
		}

		if (pending_work & SCSTIPC_WORK_RESET) {
			scstipc_proc_reset_work();
		}

		if (pending_work & SCSTIPC_WORK_STOP) {
			break;
		}
	}

	while (!kthread_should_stop()) {
		msleep(10);
	}

	PRINT_INFO("%s\n", "scst_ipc_send thread exit.");

	return 0;
}

static int scstipc_create_send_thread(void)
{
	scstipc_send_thread = kthread_create(scstipc_do_send_thread, NULL, "scst_ipc_send");
	if (IS_ERR(scstipc_send_thread)) {
		printk(KERN_ERR "kthread_create() failed.\n");
		return PTR_ERR(scstipc_do_send_thread);
	}

	return 0;
}

int scstipc_init_delayed_work(void)
{
	spin_lock_init(&scstipc_lock);
	init_waitqueue_head(&scstipc_waitqueue);

	scstipc_create_send_thread();

	PRINT_INFO("%s\n", "scst_ipc_send thread start.");

	wake_up_process(scstipc_send_thread);

	return 0;
}

void scstipc_exit_delayed_work(void)
{
	if (scstipc_send_thread) {

		scstipc_set_pending_work(SCSTIPC_WORK_STOP);

		kthread_stop(scstipc_send_thread);
	}

	return;
}

/* end of scst_ipc_delayed_work.c */
