/*
 *  scst_pcimem_drv.c
 *
 *  Copyright (C) 2004 - 2011 Vladislav Bolkhovitin <vst@vlnb.net>
 *  Copyright (C) 2004 - 2005 Leonid Stoljar
 *  Copyright (C) 2007 Ming Zhang <blackmagic02881 at gmail dot com>
 *  Copyright (C) 2007 Ross Walker <rswwalker at hotmail dot com>
 *  Copyright (C) 2007 - 2010 ID7 Ltd.
 *  Copyright (C) 2010 - 2011 SCST Ltd.
 *
 *  SCSI disk (type 0) dev handler using PCI-memory space (PCIMEM).
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, version 2
 *  of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>

#ifdef INSIDE_KERNEL_TREE
#include <scst/scst.h>
#else
#include "scst.h"
#endif

#include "scst_dev_handler.h"
#include "scst_pcimem.h"

#include "../../../../sony/ods_fpga/include/ods_fpga.h"

#if defined(CONFIG_SCST_DEBUG) || defined(CONFIG_SCST_TRACING)

static struct scst_trace_log pcimem_local_trace_tbl[] = {
	{ TRACE_ORDER,		"order" },
	{ 0,			NULL }
};
#define trace_log_tbl			pcimem_local_trace_tbl

#define PCIMEM_TRACE_TBL_HELP	", order"

#endif /* defined(CONFIG_SCST_DEBUG) || defined(CONFIG_SCST_TRACING) */

static struct ods_fpga_region pcimem_region[ODS_FPGA_REGION_MAX];
static int pcimem_num_region;

static int pcimem_attach(struct scst_device *dev);
static void pcimem_detach(struct scst_device *dev);
static int pcimem_attach_tgt(struct scst_tgt_dev *tgt_dev);
static void pcimem_detach_tgt(struct scst_tgt_dev *tgt_dev);
static ssize_t pcimem_add_device(const char *device_name, char *params);
static ssize_t pcimem_del_device(const char *device_name);


static ssize_t pcimem_sysfs_size_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf);
static ssize_t pcimem_sysfs_blocksize_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf);
static ssize_t pcimem_sysfs_rd_only_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf);
static ssize_t pcimem_sysfs_removable_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf);
static ssize_t pcimem_sysfs_media_exists_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf);
static ssize_t pcimem_sysfs_ibmbxid_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf);
static ssize_t pcimem_sysfs_obmbxid_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf);
static ssize_t pcimem_sysfs_cmdstatus_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf);
static ssize_t pcimem_sysfs_no_reserve_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf);
static ssize_t pcimem_sysfs_exmount_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf);

static ssize_t pcimem_sysfs_media_exists_store(struct kobject *kobj,
	struct kobj_attribute *attr, const char *buf, size_t count);


static struct kobj_attribute pcimem_size_attr =
	__ATTR(size_mb, S_IRUGO, pcimem_sysfs_size_show, NULL);
static struct kobj_attribute pcimem_blocksize_attr =
	__ATTR(blocksize, S_IRUGO, pcimem_sysfs_blocksize_show, NULL);
static struct kobj_attribute pcimem_rd_only_attr =
	__ATTR(read_only, S_IRUGO, pcimem_sysfs_rd_only_show, NULL);
static struct kobj_attribute pcimem_removable_attr =
	__ATTR(removable, S_IRUGO, pcimem_sysfs_removable_show, NULL);
static struct kobj_attribute pcimem_media_exists_attr =
	__ATTR(media_exists, S_IRUGO|S_IWUSR, pcimem_sysfs_media_exists_show,
		   pcimem_sysfs_media_exists_store);
static struct kobj_attribute pcimem_ibmbxid_attr =
	__ATTR(ibmbxid, S_IRUGO, pcimem_sysfs_ibmbxid_show, NULL);
static struct kobj_attribute pcimem_obmbxid_attr =
	__ATTR(obmbxid, S_IRUGO, pcimem_sysfs_obmbxid_show, NULL);
static struct kobj_attribute pcimem_cmdstatus_attr =
	__ATTR(cmdstatus, S_IRUGO, pcimem_sysfs_cmdstatus_show, NULL);
static struct kobj_attribute pcimem_no_reserve_attr =
	__ATTR(no_reserve, S_IRUGO, pcimem_sysfs_no_reserve_show, NULL);
static struct kobj_attribute pcimem_exmount_attr =
	__ATTR(exmount, S_IRUGO, pcimem_sysfs_exmount_show, NULL);


static const struct attribute *pcimem_attrs[] = {
	&pcimem_size_attr.attr,
	&pcimem_blocksize_attr.attr,
	&pcimem_rd_only_attr.attr,
	&pcimem_removable_attr.attr,
	&pcimem_media_exists_attr.attr,
	&pcimem_ibmbxid_attr.attr,
	&pcimem_obmbxid_attr.attr,
	&pcimem_cmdstatus_attr.attr,
	&pcimem_no_reserve_attr.attr,
	&pcimem_exmount_attr.attr,
	NULL,
};

/* Protects vdisks addition/deletion and related activities, like search */
static DEFINE_MUTEX(scst_pcimem_mutex);

/* Protected by scst_vdisk_mutex */
static LIST_HEAD(pcimem_dev_list);


static struct scst_dev_type pcimem_devtype = {
	.name =			"pcimem",
	.type =			TYPE_DISK,
	.threads_num =		2,
	.parse_atomic =		1,
	.dev_done_atomic =	1,
	.vendor_unique_cmd_enable = 1,
	.silent_check_blocked_dev = 1,
#ifdef CONFIG_SCST_PROC
	.no_proc =		1,
#endif
	.attach =		pcimem_attach,
	.detach =		pcimem_detach,
	.attach_tgt =		pcimem_attach_tgt,
	.detach_tgt =		pcimem_detach_tgt,
	.parse =		pcimem_parse,
	.exec =			pcimem_do_job,
#ifdef USE_PCIMEM_DUMMY
	.alloc_data_buf =	pcimem_alloc_data_buf,
	.on_free_cmd =		pcimem_on_free_cmd,
#endif /* USE_PCIMEM_DUMMY */
	.task_mgmt_fn =		pcimem_task_mgmt_fn,
#ifndef CONFIG_SCST_PROC
	.add_device =		pcimem_add_device,
	.del_device =		pcimem_del_device,
	.dev_attrs =		pcimem_attrs,
	.add_device_parameters = "blocksize, read_only, removable, media_exists, ibmbxid, obmbxid, devtype, no_reserve, exmount",
#endif
#if defined(CONFIG_SCST_DEBUG) || defined(CONFIG_SCST_TRACING)
	.default_trace_flags =	SCST_DEFAULT_DEV_LOG_FLAGS,
	.trace_flags =		&trace_flag,
	.trace_tbl =		pcimem_local_trace_tbl,
#ifndef CONFIG_SCST_PROC
	.trace_tbl_help =	PCIMEM_TRACE_TBL_HELP,
#endif
#endif
};


/* scst_pcimem_mutex supposed to be held */
static struct scst_pcimem_dev *dev_find(const char *name)
{
	struct scst_pcimem_dev *res, *vv;

	TRACE_ENTRY();

	res = NULL;
	list_for_each_entry(vv, &pcimem_dev_list, dev_list_entry) {
		if (strcmp(vv->name, name) == 0) {
			res = vv;
			break;
		}
	}

	TRACE_EXIT_HRES((unsigned long)res);
	return res;
}

static int pcimem_attach(struct scst_device *dev)
{
	int res = 0;
	struct scst_pcimem_dev *virt_dev;
	int i;

	TRACE_ENTRY();

	TRACE_DBG("virt_id %d (%s)", dev->virt_id, dev->virt_name);

	if (dev->virt_id == 0) {
		PRINT_ERROR("%s", "Not a virtual device");
		res = -EINVAL;
		goto out;
	}

	virt_dev = dev_find(dev->virt_name);
	if (virt_dev == NULL) {
		PRINT_ERROR("Device %s not found", dev->virt_name);
		res = -EINVAL;
		goto out;
	}

	virt_dev->dev = dev;

	dev->rd_only = virt_dev->rd_only;

	virt_dev->file_size = 0;
	for (i = 0; i < pcimem_num_region; i++) {
		virt_dev->file_size += pcimem_region[i].len;
	}
	virt_dev->nblocks = virt_dev->file_size >> virt_dev->block_shift;

	PRINT_INFO("Attached SCSI target virtual DISK %s "
			   "(size=%lldMB, bs=%d, nblocks=%lld,"
			   " cyln=%lld%s)",
			   virt_dev->name,
			   virt_dev->file_size >> 20, virt_dev->block_size,
			   (long long unsigned int)virt_dev->nblocks,
			   (long long unsigned int)virt_dev->nblocks/64/32,
			   virt_dev->nblocks < 64*32
			   ? " !WARNING! cyln less than 1" : "");

	dev->dh_priv = virt_dev;

	dev->tst = DEF_TST;
	dev->d_sense = DEF_DSENSE;
	if (virt_dev->wt_flag && !virt_dev->nv_cache)
		dev->queue_alg = DEF_QUEUE_ALG_WT;
	else
		dev->queue_alg = DEF_QUEUE_ALG;
	dev->swp = DEF_SWP;
	dev->tas = DEF_TAS;

out:
	TRACE_EXIT();
	return res;
}

static void pcimem_detach(struct scst_device *dev)
{
	TRACE_ENTRY();

	TRACE_DBG("virt_id %d", dev->virt_id);

	PRINT_INFO("Detached virtual device %s", dev->virt_name);

	/* virt_dev will be freed by the caller */
	dev->dh_priv = NULL;

	TRACE_EXIT();
	return;
}

static int pcimem_attach_tgt(struct scst_tgt_dev *tgt_dev)
{
	int res = 0;

	TRACE_ENTRY();

	/* Nothing to do */

	TRACE_EXIT_RES(res);
	return res;
}

static void pcimem_detach_tgt(struct scst_tgt_dev *tgt_dev)
{
	TRACE_ENTRY();

	scst_del_all_thr_data(tgt_dev);

	TRACE_EXIT();
	return;
}

static int pcimem_parse_add_dev_params(struct scst_pcimem_dev *virt_dev,
	char *params, const char *allowed_params[])
{
	int res = 0;
	unsigned long val;
	char *param, *p, *pp;

	TRACE_ENTRY();

	while (1) {
		param = scst_get_next_token_str(&params);
		if (param == NULL)
			break;

		p = scst_get_next_lexem(&param);
		if (*p == '\0') {
			PRINT_ERROR("Syntax error at %s (device %s)",
				param, virt_dev->name);
			res = -EINVAL;
			goto out;
		}

		if (allowed_params != NULL) {
			const char **a = allowed_params;
			bool allowed = false;

			while (*a != NULL) {
				if (!strcasecmp(*a, p)) {
					allowed = true;
					break;
				}
				a++;
			}

			if (!allowed) {
				PRINT_ERROR("Unknown parameter %s (device %s)", p,
					virt_dev->name);
				res = -EINVAL;
				goto out;
			}
		}

		pp = scst_get_next_lexem(&param);
		if (*pp == '\0') {
			PRINT_ERROR("Parameter %s value missed for device %s",
				p, virt_dev->name);
			res = -EINVAL;
			goto out;
		}

		if (scst_get_next_lexem(&param)[0] != '\0') {
			PRINT_ERROR("Too many parameter's %s values (device %s)",
				p, virt_dev->name);
			res = -EINVAL;
			goto out;
		}

#if 0
		if (!strcasecmp("filename", p)) {
			if (*pp != '/') {
				PRINT_ERROR("Filename %s must be global "
					"(device %s)", pp, virt_dev->name);
				res = -EINVAL;
				goto out;
			}

			virt_dev->filename = kstrdup(pp, GFP_KERNEL);
			if (virt_dev->filename == NULL) {
				PRINT_ERROR("Unable to duplicate file name %s "
					"(device %s)", pp, virt_dev->name);
				res = -ENOMEM;
				goto out;
			}
			continue;
		}
#endif

		res = strict_strtoul(pp, 0, &val);
		if (res != 0) {
			PRINT_ERROR("strict_strtoul() for %s failed: %d "
				"(device %s)", pp, res, virt_dev->name);
			goto out;
		}

		if (!strcasecmp("write_through", p)) {
			virt_dev->wt_flag = val;
			TRACE_DBG("WRITE THROUGH %d", virt_dev->wt_flag);
		} else if (!strcasecmp("nv_cache", p)) {
			virt_dev->nv_cache = val;
			TRACE_DBG("NON-VOLATILE CACHE %d", virt_dev->nv_cache);
		} else if (!strcasecmp("o_direct", p)) {
#if 0
			virt_dev->o_direct_flag = val;
			TRACE_DBG("O_DIRECT %d", virt_dev->o_direct_flag);
#else
			PRINT_INFO("O_DIRECT flag doesn't currently"
				" work, ignoring it, use fileio_tgt "
				"in O_DIRECT mode instead (device %s)", virt_dev->name);
#endif
		} else if (!strcasecmp("read_only", p)) {
			virt_dev->rd_only = val;
			TRACE_DBG("READ ONLY %d", virt_dev->rd_only);
		} else if (!strcasecmp("removable", p)) {
			virt_dev->removable = val;
			TRACE_DBG("REMOVABLE %d", virt_dev->removable);
		} else if (!strcasecmp("thin_provisioned", p)) {
			virt_dev->thin_provisioned = val;
			TRACE_DBG("THIN PROVISIONED %d",
				virt_dev->thin_provisioned);
		} else if (!strcasecmp("blocksize", p)) {
			virt_dev->block_size = val;
			virt_dev->block_shift = scst_calc_block_shift(
							virt_dev->block_size);
			if (virt_dev->block_shift < 9) {
				res = -EINVAL;
				goto out;
			}
			TRACE_DBG("block_size %d, block_shift %d",
				virt_dev->block_size,
				virt_dev->block_shift);
		} else if (!strcasecmp("media_exists", p)) {
			virt_dev->media_exists = val;
			TRACE_DBG("media_exists %d", virt_dev->media_exists);
		} else if (!strcasecmp("ibmbxid", p)) {
			virt_dev->ibmbxid = val;
			TRACE_DBG("ibmbxid %d", virt_dev->ibmbxid);
		} else if (!strcasecmp("obmbxid", p)) {
			virt_dev->obmbxid = val;
			TRACE_DBG("obmbxid %d", virt_dev->obmbxid);
		} else if (!strcasecmp("devtype", p)) {
			virt_dev->devtype = val;
			TRACE_DBG("devtype %d", virt_dev->devtype);
		} else if (!strcasecmp("no_reserve", p)) {
			virt_dev->no_reserve = val;
			TRACE_DBG("no_reserve %d", virt_dev->no_reserve);
		} else if (!strcasecmp("exmount", p)) {
			virt_dev->exmount = val;
			TRACE_DBG("exmount %d", virt_dev->exmount);
		} else {
			PRINT_ERROR("Unknown parameter %s (device %s)", p,
				virt_dev->name);
			res = -EINVAL;
			goto out;
		}
	}

out:
	TRACE_EXIT_RES(res);
	return res;
}

static ssize_t pcimem_add_device(const char *device_name, char *params)
{
	int res;
	const char *allowed_params[] = { "read_only", "removable", "blocksize", "media_exists", "ibmbxid", "obmbxid", "devtype", "no_reserve", "exmount", NULL };
	struct scst_pcimem_dev *virt_dev;

	TRACE_ENTRY();

	res = -EEXIST;
	if (dev_find(device_name)) {
		goto out;
	}

	virt_dev = kzalloc(sizeof(*virt_dev), GFP_KERNEL);
	if (virt_dev == NULL) {
		PRINT_ERROR("Allocation of virtual device %s failed",
			device_name);
		res = -ENOMEM;
		goto out;
	}

	sema_init(&virt_dev->mbx_sem, 1);

	if (strlen(device_name) >= sizeof(virt_dev->name)) {
		PRINT_ERROR("Name %s is too long (max allowed %zd)", device_name,
			sizeof(virt_dev->name)-1);
		res = -EINVAL;
		goto out_free;
	}
	strcpy(virt_dev->name, device_name);

	pcimem_add_virtual_device(virt_dev);

	res = pcimem_parse_add_dev_params(virt_dev, params, allowed_params);
	if (res != 0)
		goto out_free;

	if (mutex_lock_interruptible(&scst_pcimem_mutex) != 0) {
		res = -EINTR;
		goto out_free;
	}

	pcimem_devtype.type = virt_dev->devtype;

	list_add_tail(&virt_dev->dev_list_entry, &pcimem_dev_list);

	virt_dev->virt_id = scst_register_virtual_device(&pcimem_devtype, device_name);

	pcimem_devtype.type = TYPE_DISK;

	mutex_unlock(&scst_pcimem_mutex);

	if (virt_dev->virt_id < 0) {
		res = virt_dev->virt_id;
		goto out_free;
	}

	TRACE_DBG("Registered virt_dev %s with id %d", device_name, virt_dev->virt_id);

out:
	TRACE_EXIT_RES(res);
	return res;

out_free:
	kfree(virt_dev);
	goto out;
}

static ssize_t pcimem_del_device(const char *device_name)
{
	int res = 0;
	struct scst_pcimem_dev *virt_dev;

	TRACE_ENTRY();

	if (mutex_lock_interruptible(&scst_pcimem_mutex) != 0) {
		res = -EINTR;
		goto out;
	}

	virt_dev = dev_find(device_name);
	if (virt_dev == NULL) {
		PRINT_ERROR("Device %s not found", device_name);
		res = -EINVAL;
		goto out_unlock;
	}

	scst_unregister_virtual_device(virt_dev->virt_id);

	list_del(&virt_dev->dev_list_entry);

	pcimem_del_virtual_device(virt_dev);

	PRINT_INFO("Virtual device %s unregistered", device_name);
	TRACE_DBG("virt_id %d unregistered", virt_dev->virt_id);

	kfree(virt_dev);

out_unlock:
	mutex_unlock(&scst_pcimem_mutex);

out:
	TRACE_EXIT_RES(res);
	return res;
}

int pcimem_get_block_shift(struct scst_cmd *cmd)
{
	struct scst_pcimem_dev *virt_dev = cmd->dev->dh_priv;
	return virt_dev->block_shift;
}

static ssize_t pcimem_sysfs_size_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf)
{
	int pos = 0;
	struct scst_device *dev;
	struct scst_pcimem_dev *virt_dev;

	TRACE_ENTRY();

	dev = container_of(kobj, struct scst_device, dev_kobj);
	virt_dev = dev->dh_priv;

	pos = sprintf(buf, "%lld\n", virt_dev->file_size / 1024 / 1024);

	TRACE_EXIT_RES(pos);
	return pos;
}

static ssize_t pcimem_sysfs_blocksize_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf)
{
	int pos = 0;
	struct scst_device *dev;
	struct scst_pcimem_dev *virt_dev;

	TRACE_ENTRY();

	dev = container_of(kobj, struct scst_device, dev_kobj);
	virt_dev = dev->dh_priv;

	pos = sprintf(buf, "%d\n%s", (int)virt_dev->block_size,
		(virt_dev->block_size == DEF_DISK_BLOCKSIZE) ? "" :
			SCST_SYSFS_KEY_MARK "\n");

	TRACE_EXIT_RES(pos);
	return pos;
}

static ssize_t pcimem_sysfs_rd_only_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf)
{
	int pos = 0;
	struct scst_device *dev;
	struct scst_pcimem_dev *virt_dev;

	TRACE_ENTRY();

	dev = container_of(kobj, struct scst_device, dev_kobj);
	virt_dev = dev->dh_priv;

	pos = sprintf(buf, "%d\n%s", virt_dev->rd_only ? 1 : 0,
		(virt_dev->rd_only == DEF_RD_ONLY) ? "" :
			SCST_SYSFS_KEY_MARK "\n");

	TRACE_EXIT_RES(pos);
	return pos;
}

static ssize_t pcimem_sysfs_removable_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf)
{
	int pos = 0;
	struct scst_device *dev;
	struct scst_pcimem_dev *virt_dev;

	TRACE_ENTRY();

	dev = container_of(kobj, struct scst_device, dev_kobj);
	virt_dev = dev->dh_priv;

	pos = sprintf(buf, "%d\n", virt_dev->removable ? 1 : 0);

	if ((virt_dev->dev->type != TYPE_ROM) &&
	    (virt_dev->removable != DEF_REMOVABLE))
		pos += sprintf(&buf[pos], "%s\n", SCST_SYSFS_KEY_MARK);

	TRACE_EXIT_RES(pos);
	return pos;
}

static ssize_t pcimem_sysfs_media_exists_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf)
{
	int pos = 0;
	struct scst_device *dev;
	struct scst_pcimem_dev *virt_dev;

	TRACE_ENTRY();

	dev = container_of(kobj, struct scst_device, dev_kobj);
	virt_dev = dev->dh_priv;

	pos = sprintf(buf, "%d\n", virt_dev->media_exists ? 1 : 0);

	if (virt_dev->media_exists != DEF_MEDIA_EXISTS) {
		pos += sprintf(&buf[pos], "%s\n", SCST_SYSFS_KEY_MARK);
	}

	TRACE_EXIT_RES(pos);
	return pos;
}

static ssize_t pcimem_sysfs_ibmbxid_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf)
{
	int pos = 0;
	struct scst_device *dev;
	struct scst_pcimem_dev *virt_dev;

	TRACE_ENTRY();

	dev = container_of(kobj, struct scst_device, dev_kobj);
	virt_dev = dev->dh_priv;

	pos = sprintf(buf, "%d\n", virt_dev->ibmbxid);

	TRACE_EXIT_RES(pos);
	return pos;
}

static ssize_t pcimem_sysfs_obmbxid_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf)
{
	int pos = 0;
	struct scst_device *dev;
	struct scst_pcimem_dev *virt_dev;

	TRACE_ENTRY();

	dev = container_of(kobj, struct scst_device, dev_kobj);
	virt_dev = dev->dh_priv;

	pos = sprintf(buf, "%d\n", virt_dev->obmbxid);

	TRACE_EXIT_RES(pos);
	return pos;
}

static int pcimem_sysfs_process_media_exists_store(struct scst_sysfs_work_item *work)
{
	int res = 0;
	struct scst_device *dev = work->dev;
	struct scst_pcimem_dev *virt_dev;
	int media_exists;

	TRACE_ENTRY();

	/* It's safe, since we taken dev_kobj and dh_priv NULLed in attach() */
	virt_dev = dev->dh_priv;

	switch (work->buf[0]) {
	case '0':
		media_exists = 0;
		break;
	case '1':
		media_exists = 1;
		break;
	default:
		res = -EINVAL;
		goto out;
	}

	mutex_lock(&scst_mutex);

	virt_dev->media_exists = media_exists;
	if (virt_dev->media_exists) {
		virt_dev->media_changed = 1;
	}

	mutex_unlock(&scst_mutex);

	kobject_put(&dev->dev_kobj);

out:
	TRACE_EXIT_RES(res);
	return res;
}

static ssize_t pcimem_sysfs_media_exists_store(struct kobject *kobj,
	struct kobj_attribute *attr, const char *buf, size_t count)
{
	int res;
	char *i_buf;
	struct scst_sysfs_work_item *work;
	struct scst_device *dev;

	TRACE_ENTRY();

	dev = container_of(kobj, struct scst_device, dev_kobj);

	i_buf = kasprintf(GFP_KERNEL, "%.*s", (int)count, buf);
	if (i_buf == NULL) {
		PRINT_ERROR("Unable to alloc intermediate buffer with size %zd",
			count+1);
		res = -ENOMEM;
		goto out;
	}

	res = scst_alloc_sysfs_work(pcimem_sysfs_process_media_exists_store,
					false, &work);
	if (res != 0)
		goto out_free;

	work->buf = i_buf;
	work->dev = dev;

	kobject_get(&dev->dev_kobj);

	res = scst_sysfs_queue_wait_work(work);
	if (res == 0)
		res = count;

out:
	TRACE_EXIT_RES(res);
	return res;

out_free:
	kfree(i_buf);
	goto out;
}

static ssize_t pcimem_sysfs_cmdstatus_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf)
{
	int pos = 0;
	struct scst_device *dev;
	struct scst_pcimem_dev *virt_dev;
	struct scst_cmd *active_cmd;
	uint8_t cdb[SCST_MAX_CDB_SIZE];
	scst_data_direction data_direction;
	int bufflen;
	int state;
	const char *initiator_name;
	unsigned long elapsed;
	unsigned short curr_sn;

	TRACE_ENTRY();

	dev = container_of(kobj, struct scst_device, dev_kobj);
	virt_dev = dev->dh_priv;

	spin_lock(&virt_dev->active_cmdinfo_lock);
	active_cmd = virt_dev->ipc_active_cmd;
	if (active_cmd) {
		memcpy(cdb, active_cmd->cdb, sizeof(cdb));
		data_direction = active_cmd->data_direction;
		bufflen = active_cmd->bufflen;
		state = active_cmd->state;
		initiator_name = active_cmd->sess->initiator_name;
	}
	curr_sn = virt_dev->curr_sn;
	elapsed = jiffies - virt_dev->last_updated;
	spin_unlock(&virt_dev->active_cmdinfo_lock);

	pos += sprintf(&buf[pos], "cmd: %p\n", active_cmd);

	if (active_cmd) {
		pos += sprintf(&buf[pos], "cdb: %02x %02x %02x %02x ", cdb[0], cdb[1], cdb[2], cdb[3]);
		pos += sprintf(&buf[pos], " %02x %02x %02x %02x ", cdb[4], cdb[5], cdb[6], cdb[7]);
		pos += sprintf(&buf[pos], " %02x %02x %02x %02x ", cdb[8], cdb[9], cdb[10], cdb[11]);
		pos += sprintf(&buf[pos], " %02x %02x %02x %02x \n", cdb[12], cdb[13], cdb[14], cdb[15]);

		pos += sprintf(&buf[pos], "data_direction: %d\n", data_direction);
		pos += sprintf(&buf[pos], "bufflen: %d\n", bufflen);
		pos += sprintf(&buf[pos], "state: %d\n", state);
		pos += sprintf(&buf[pos], "initiator_name: %s\n", initiator_name);
	}

	pos += sprintf(&buf[pos], "num cmd: %u\n", curr_sn);
	pos += sprintf(&buf[pos], "elapsed time: %lu [ms]\n", elapsed * 1000 / HZ);

	TRACE_EXIT_RES(pos);
	return pos;
}

static ssize_t pcimem_sysfs_no_reserve_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf)
{
	int pos = 0;
	struct scst_device *dev;
	struct scst_pcimem_dev *virt_dev;

	TRACE_ENTRY();

	dev = container_of(kobj, struct scst_device, dev_kobj);
	virt_dev = dev->dh_priv;

	pos = sprintf(buf, "%d\n", virt_dev->no_reserve);

	TRACE_EXIT_RES(pos);
	return pos;
}

static ssize_t pcimem_sysfs_exmount_show(struct kobject *kobj,
	struct kobj_attribute *attr, char *buf)
{
	int pos = 0;
	struct scst_device *dev;
	struct scst_pcimem_dev *virt_dev;

	TRACE_ENTRY();

	dev = container_of(kobj, struct scst_device, dev_kobj);
	virt_dev = dev->dh_priv;

	pos = sprintf(buf, "%d\n", virt_dev->exmount);

	TRACE_EXIT_RES(pos);
	return pos;
}

unsigned long *get_trace_flag_ptr(void)
{
	return &trace_flag;
}

unsigned long pcimem_bus_to_phys(unsigned long address)
{
	int i;

	for (i = 0; i < pcimem_num_region; i++) {	
		if ((pcimem_region[i].bus_base <= address) &&
			(address < pcimem_region[i].bus_base + pcimem_region[i].len)) {
			address -= pcimem_region[i].bus_base;
			address += pcimem_region[i].phys_base;
		}
	}

	return address;
}

int pcimem_add_sg_entries(struct scatterlist *sg, int num_sg, unsigned long offset, unsigned long size)
{
	unsigned long region_top = 0;
	unsigned long added_size = 0;
	unsigned long pos = offset;
	unsigned long region_offset;
	unsigned long add_len;
	int sg_index = 0;
	int s = 0;
   	int i;

	while (added_size < size) {
		if (sg_index >= num_sg) {
			PRINT_ERROR("too many sg(%d >= %d)\n", sg_index, num_sg);
			return -1;
		}

		for (i = s; i < pcimem_num_region; i++) {
			if ((pos >= region_top) && (pos < region_top + pcimem_region[i].len)) {
				break;
			}

			region_top += pcimem_region[i].len;
		}

		if (i == pcimem_num_region) {
			PRINT_ERROR("no pci_region found(offset=0x%08lx,size=0x%08lx)\n", offset, size);
			return -1;
		}

		region_offset = pos - region_top;
		add_len = (size - added_size) < (pcimem_region[i].len - region_offset) ? (size - added_size) : (pcimem_region[i].len - region_offset);

		sg[sg_index].dma_address = pcimem_region[i].bus_base
			+ region_offset;
		sg[sg_index].dma_length = add_len;
		sg_index++;

		pos += add_len;
		added_size += add_len;
		region_top += pcimem_region[i].len;
		s = i + 1;
	}

	return sg_index;
}

static int __init init_scst_pcimem_driver(void)
{
	int res = 0;
	int i;

	TRACE_ENTRY();

	pcimem_num_region = ods_fpga_get_num_dram_region();
	if (pcimem_num_region == 0) {
		res = -ENODEV;
		goto out;
	}

	for (i = 0; i < pcimem_num_region; i++) {
		if (ods_fpga_get_region_info(ODS_FPGA_REGION_DRAM0 + i, &pcimem_region[i]) != 0) {
			res = -ENODEV;
			goto out;
		}

		PRINT_INFO("region[%d] phys_base=0x%08lx,bus_base=0x%08lx,len=0x%lx", i, pcimem_region[i].phys_base, pcimem_region[i].bus_base, pcimem_region[i].len);
	}

	pcimem_init_virtual_dev();

	res = scst_register_virtual_dev_driver(&pcimem_devtype);
	if (res < 0) {
		goto out;
	}

out:
	TRACE_EXIT_RES(res);
	return res;	

}

static void __exit exit_scst_pcimem_driver(void)
{
	TRACE_ENTRY();

	scst_unregister_virtual_dev_driver(&pcimem_devtype);

	pcimem_exit_virtual_dev();

	TRACE_EXIT();
	return;
}

module_init(init_scst_pcimem_driver);
module_exit(exit_scst_pcimem_driver);
MODULE_LICENSE("GPL");

/* end of scst_pcimem_drv.c */
