/*
 *  scst_pcimem_dummy.c
 *
 *  Copyright (C) 2004 - 2011 Vladislav Bolkhovitin <vst@vlnb.net>
 *  Copyright (C) 2004 - 2005 Leonid Stoljar
 *  Copyright (C) 2007 Ming Zhang <blackmagic02881 at gmail dot com>
 *  Copyright (C) 2007 Ross Walker <rswwalker at hotmail dot com>
 *  Copyright (C) 2007 - 2010 ID7 Ltd.
 *  Copyright (C) 2010 - 2011 SCST Ltd.
 *
 *  SCSI disk (type 0) dev handler using PCI-memory space (PCIMEM).
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, version 2
 *  of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/crc32c.h>

#include <asm/unaligned.h>

#ifdef INSIDE_KERNEL_TREE
#include <scst/scst.h>
#else
#include "scst.h"
#endif

#ifdef USE_IPCIF
#include "../../../../sony/ipcif_scstlib/ipcif_scstlib.h"
#endif /* USE_IPCIF */

#include "scst_dev_handler.h"
#include "scst_mem.h"
#include "scst_pcimem.h"

#if defined(CONFIG_SCST_DEBUG) || defined(CONFIG_SCST_TRACING)

static unsigned long *p_dh_trace_flag;

#undef trace_flag
#define trace_flag (*p_dh_trace_flag)

#endif /* defined(CONFIG_SCST_DEBUG) || defined(CONFIG_SCST_TRACING) */


/* 8 byte ASCII Vendor */
#define SCST_FIO_VENDOR			"SCST_FIO"
#define SCST_BIO_VENDOR			"SCST_BIO"
/* 4 byte ASCII Product Revision Level - left aligned */
#define SCST_FIO_REV			" 220"

#define INQ_BUF_SZ			256
#define EVPD				0x01
#define CMDDT				0x02

#define MSENSE_BUF_SZ			256
#define DBD				0x08	/* disable block descriptor */
#define WP				0x80	/* write protect */
#define DPOFUA				0x10	/* DPOFUA bit */
#define WCE				0x04	/* write cache enable */

#define PF				0x10	/* page format */
#define SP				0x01	/* save pages */
#define PS				0x80	/* parameter saveable */

#define	BYTE				8
#define	DEF_SECTORS			56
#define	DEF_HEADS			255

#define MAX_SGV_CNT 4

/* Protects devices t10_dev_id and usn */
static DEFINE_RWLOCK(pcimem_serial_rwlock);


#ifdef USE_IPCIF
static void pcimem_proc_cmdresp(void *priv, unsigned int tag, unsigned int resp_code, unsigned int sc, unsigned int asc, unsigned int ascq, unsigned int add_inf, unsigned int add_spec);
static void pcimem_proc_dmardy(void *priv, unsigned int tag, struct ipcscst_areainfo *area, int num_area, int direction);
static void pcimem_proc_reset_resp(void *priv);
static int pcimem_send_cmdreq(struct scst_cmd *cmd);

static struct ipcscst_funcs ipcscst_funcs = {
	pcimem_proc_cmdresp,
	pcimem_proc_dmardy,
	pcimem_proc_reset_resp,
};
#endif /* USE_IPCIF */

static uint64_t gen_dev_id_num(const char *virt_dev_name)
{
	uint32_t dev_id_num;

	dev_id_num = crc32c(0, virt_dev_name, strlen(virt_dev_name)+1);

#ifdef CONFIG_SCST_PROC
	return ((uint64_t)scst_vdisk_ID << 32) | dev_id_num;
#else
	return ((uint64_t)scst_get_setup_id() << 32) | dev_id_num;
#endif
}

int pcimem_parse(struct scst_cmd *cmd)
{
	scst_sbc_generic_parse(cmd, pcimem_get_block_shift);
	return SCST_CMD_STATE_DEFAULT;
}

static void pcimem_exec_unmap(struct scst_cmd *cmd)
{
	struct scst_pcimem_dev *virt_dev = cmd->dev->dh_priv;
	//	ssize_t length = 0;
	//	struct file *fd = thr->fd;
	//	struct inode *inode;
	//	uint8_t *address;
	//	int offset, descriptor_len, total_len;

	TRACE_ENTRY();

	if (unlikely(!virt_dev->thin_provisioned)) {
		TRACE_DBG("%s", "Invalid opcode UNMAP");
		scst_set_cmd_error(cmd,
			SCST_LOAD_SENSE(scst_sense_invalid_opcode));
		goto out;
	}

	if (unlikely(cmd->cdb[1] & 1)) {
		/* ANCHOR not supported */
		TRACE_DBG("%s", "Invalid ANCHOR field");
		scst_set_cmd_error(cmd,
			SCST_LOAD_SENSE(scst_sense_invalid_field_in_cdb));
		goto out;
	}

#if 0
	length = scst_get_buf_full(cmd, &address);
	if (unlikely(length <= 0)) {
		if (length == 0)
			goto out_put;
		else if (length == -ENOMEM)
			scst_set_busy(cmd);
		else
			scst_set_cmd_error(cmd,
				SCST_LOAD_SENSE(scst_sense_hardw_error));
		goto out;
	}

	inode = fd->f_dentry->d_inode;

	total_len = cmd->cdb[7] << 8 | cmd->cdb[8]; /* length */
	offset = 8;

	descriptor_len = address[2] << 8 | address[3];

	TRACE_DBG("total_len %d, descriptor_len %d", total_len, descriptor_len);

	if (descriptor_len == 0)
		goto out_put;

	if (unlikely((descriptor_len > (total_len - 8)) ||
		     ((descriptor_len % 16) != 0))) {
		PRINT_ERROR("Bad descriptor length: %d < %d - 8",
			descriptor_len, total_len);
		scst_set_cmd_error(cmd,
			SCST_LOAD_SENSE(scst_sense_invalid_field_in_parm_list));
		goto out_put;
	}

	while ((offset - 8) < descriptor_len) {
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 27)
		int err;
#endif
		uint64_t start;
		uint32_t len;
		start = be64_to_cpu(get_unaligned((__be64 *)&address[offset]));
		offset += 8;
		len = be32_to_cpu(get_unaligned((__be32 *)&address[offset]));
		offset += 8;

		if ((start > virt_dev->nblocks) ||
		    ((start + len) > virt_dev->nblocks)) {
			PRINT_ERROR("Device %s: attempt to write beyond max "
				"size", virt_dev->name);
			scst_set_cmd_error(cmd,
			    SCST_LOAD_SENSE(scst_sense_invalid_field_in_cdb));
			goto out_put;
		}

		TRACE_DBG("Unmapping lba %lld (blocks %d)",
			(unsigned long long)start, len);

		if (virt_dev->blockio) {
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 27)
			gfp_t gfp = cmd->noio_mem_alloc ? GFP_NOIO : GFP_KERNEL;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 31)
			err = blkdev_issue_discard(inode->i_bdev, start, len,
					gfp);
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 35)       \
      && !(LINUX_VERSION_CODE == KERNEL_VERSION(2, 6, 34) \
           && defined(CONFIG_SUSE_KERNEL))
			err = blkdev_issue_discard(inode->i_bdev, start, len,
					gfp, DISCARD_FL_WAIT);
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 37)
			err = blkdev_issue_discard(inode->i_bdev, start, len,
					gfp, BLKDEV_IFL_WAIT);
#else
			err = blkdev_issue_discard(inode->i_bdev, start, len,
						   gfp, 0);
#endif
			if (unlikely(err != 0)) {
				PRINT_ERROR("blkdev_issue_discard() for "
					"LBA %lld len %d failed with err %d",
					(unsigned long long)start, len, err);
				scst_set_cmd_error(cmd,
					SCST_LOAD_SENSE(scst_sense_write_error));
				goto out_put;
			}
#else
		scst_set_cmd_error(cmd,
			SCST_LOAD_SENSE(scst_sense_invalid_opcode));
		goto out_put;
#endif
		} else {
			const int block_shift = virt_dev->block_shift;
			const loff_t a0 = start << block_shift;
			const loff_t a2 = (start + len) << block_shift;
			const loff_t a1 = max_t(loff_t, a2 & PAGE_CACHE_MASK,
						a0);

			/*
			 * The SCSI UNMAP command discards a range of blocks
			 * of size (1 << block_shift) while the Linux VFS
			 * truncate_range() function discards a range of blocks
			 * of size PAGE_CACHE_SIZE. Hence pass range [a0, a1)
			 * to truncate_range() instead of range [a0,
			 * a2). Note: since we do not set TPRZ it is not
			 * necessary to overwrite the range [a1, a2) with
			 * zeroes.
			 */
			WARN_ON(!(a0 <= a1 && a1 <= a2));
			WARN_ON(!((a1 & (PAGE_CACHE_SIZE - 1)) == 0 ||
				  a0 == a1));
			if (a0 < a1)
				inode->i_op->truncate_range(inode, a0, a1 - 1);
		}
	}

out_put:
	scst_put_buf_full(cmd, address);

#endif /* 0 */

out:
	TRACE_EXIT();
	return;
}

static void pcimem_exec_inquiry(struct scst_cmd *cmd)
{
	int32_t length, i, resp_len = 0;
	uint8_t *address;
	uint8_t *buf;
	struct scst_pcimem_dev *virt_dev = cmd->dev->dh_priv;
	uint16_t tg_id;

	TRACE_ENTRY();

	buf = kzalloc(INQ_BUF_SZ, GFP_KERNEL);
	if (buf == NULL) {
		scst_set_busy(cmd);
		goto out;
	}

	length = scst_get_buf_full(cmd, &address);
	TRACE_DBG("length %d", length);
	if (unlikely(length <= 0)) {
		if (length < 0) {
			PRINT_ERROR("scst_get_buf_full() failed: %d", length);
			scst_set_cmd_error(cmd,
				SCST_LOAD_SENSE(scst_sense_hardw_error));
		}
		goto out_free;
	}

	if (cmd->cdb[1] & CMDDT) {
		TRACE_DBG("%s", "INQUIRY: CMDDT is unsupported");
		scst_set_cmd_error(cmd,
		    SCST_LOAD_SENSE(scst_sense_invalid_field_in_cdb));
		goto out_put;
	}

	buf[0] = cmd->dev->type;      /* type dev */
	/* Vital Product */
	if (cmd->cdb[1] & EVPD) {
		if (0 == cmd->cdb[2]) {
			/* supported vital product data pages */
			buf[3] = 3;
			buf[4] = 0x0; /* this page */
			buf[5] = 0x80; /* unit serial number */
			buf[6] = 0x83; /* device identification */
			if (virt_dev->dev->type == TYPE_DISK) {
				buf[3] += 1;
				buf[7] = 0xB0; /* block limits */
				if (virt_dev->thin_provisioned) {
					buf[3] += 1;
					buf[8] = 0xB2; /* thin provisioning */
				}
			}
			resp_len = buf[3] + 4;
		} else if (0x80 == cmd->cdb[2]) {
			/* unit serial number */
			buf[1] = 0x80;
			if (cmd->tgtt->get_serial) {
				buf[3] = cmd->tgtt->get_serial(cmd->tgt_dev,
						       &buf[4], INQ_BUF_SZ - 4);
			} else {
				int usn_len;
				read_lock(&pcimem_serial_rwlock);
				usn_len = strlen(virt_dev->usn);
				buf[3] = usn_len;
				strncpy(&buf[4], virt_dev->usn, usn_len);
				read_unlock(&pcimem_serial_rwlock);
			}
			resp_len = buf[3] + 4;
		} else if (0x83 == cmd->cdb[2]) {
			/* device identification */
			int num = 4;

			buf[1] = 0x83;
			/* T10 vendor identifier field format (faked) */
			buf[num + 0] = 0x2;	/* ASCII */
			buf[num + 1] = 0x1;	/* Vendor ID */
			if (cmd->tgtt->vendor)
				memcpy(&buf[num + 4], cmd->tgtt->vendor, 8);
			else
				memcpy(&buf[num + 4], SCST_FIO_VENDOR, 8);

			read_lock(&pcimem_serial_rwlock);
			i = strlen(virt_dev->t10_dev_id);
			memcpy(&buf[num + 12], virt_dev->t10_dev_id, i);
			read_unlock(&pcimem_serial_rwlock);

			buf[num + 3] = 8 + i;
			num += buf[num + 3];

			num += 4;

			/*
			 * Relative target port identifier
			 */
			buf[num + 0] = 0x01; /* binary */
			/* Relative target port id */
			buf[num + 1] = 0x10 | 0x04;

			put_unaligned(cpu_to_be16(cmd->tgt->rel_tgt_id),
				(__be16 *)&buf[num + 4 + 2]);

			buf[num + 3] = 4;
			num += buf[num + 3];

			num += 4;

			tg_id = scst_lookup_tg_id(cmd->dev, cmd->tgt);
			if (tg_id) {
				/*
				 * Target port group designator
				 */
				buf[num + 0] = 0x01; /* binary */
				/* Target port group id */
				buf[num + 1] = 0x10 | 0x05;

				put_unaligned(cpu_to_be16(tg_id),
					      (__be16 *)&buf[num + 4 + 2]);

				buf[num + 3] = 4;
				num += 4 + buf[num + 3];
			}

			/*
			 * IEEE id
			 */
			buf[num + 0] = 0x01; /* binary */

			/* EUI-64 */
			buf[num + 1] = 0x02;
			buf[num + 2] = 0x00;
			buf[num + 3] = 0x08;

			/* IEEE id */
			buf[num + 4] = virt_dev->t10_dev_id[0];
			buf[num + 5] = virt_dev->t10_dev_id[1];
			buf[num + 6] = virt_dev->t10_dev_id[2];

			/* IEEE ext id */
			buf[num + 7] = virt_dev->t10_dev_id[3];
			buf[num + 8] = virt_dev->t10_dev_id[4];
			buf[num + 9] = virt_dev->t10_dev_id[5];
			buf[num + 10] = virt_dev->t10_dev_id[6];
			buf[num + 11] = virt_dev->t10_dev_id[7];
			num += buf[num + 3];

			resp_len = num;
			buf[2] = (resp_len >> 8) & 0xFF;
			buf[3] = resp_len & 0xFF;
			resp_len += 4;
		} else if ((0xB0 == cmd->cdb[2]) &&
			   (virt_dev->dev->type == TYPE_DISK)) {
			/* Block Limits */
			int max_transfer;
			buf[1] = 0xB0;
			buf[3] = 0x3C;
			/* Optimal transfer granuality is PAGE_SIZE */
			put_unaligned(cpu_to_be16(max_t(int,
					PAGE_SIZE/virt_dev->block_size, 1)),
				      (uint16_t *)&buf[6]);
			/* Max transfer len is min of sg limit and 8M */
			max_transfer = min_t(int,
					cmd->tgt_dev->max_sg_cnt << PAGE_SHIFT,
					8*1024*1024) / virt_dev->block_size;
			put_unaligned(cpu_to_be32(max_transfer),
					(uint32_t *)&buf[8]);
			/*
			 * Let's have optimal transfer len 512KB. Better to not
			 * set it at all, because we don't have such limit,
			 * but some initiators may not understand that (?).
			 * From other side, too big transfers  are not optimal,
			 * because SGV cache supports only <4M buffers.
			 */
			put_unaligned(cpu_to_be32(min_t(int,
					max_transfer,
					512*1024 / virt_dev->block_size)),
				      (uint32_t *)&buf[12]);
			if (virt_dev->thin_provisioned) {
				/* MAXIMUM UNMAP LBA COUNT is UNLIMITED */
				put_unaligned(__constant_cpu_to_be32(0xFFFFFFFF),
					      (uint32_t *)&buf[20]);
				/* MAXIMUM UNMAP BLOCK DESCRIPTOR COUNT is UNLIMITED */
				put_unaligned(__constant_cpu_to_be32(0xFFFFFFFF),
					      (uint32_t *)&buf[24]);
				/* OPTIMAL UNMAP GRANULARITY is 1 */
				put_unaligned(__constant_cpu_to_be32(1),
					      (uint32_t *)&buf[28]);
			}
			resp_len = buf[3] + 4;
		} else if ((0xB2 == cmd->cdb[2]) &&
			   (virt_dev->dev->type == TYPE_DISK) &&
			   virt_dev->thin_provisioned) {
			/* Thin Provisioning */
			buf[1] = 0xB2;
			buf[3] = 2;
			buf[5] = 0x80;
			resp_len = buf[3] + 4;
		} else {
			TRACE_DBG("INQUIRY: Unsupported EVPD page %x",
				cmd->cdb[2]);
			scst_set_cmd_error(cmd,
			    SCST_LOAD_SENSE(scst_sense_invalid_field_in_cdb));
			goto out_put;
		}
	} else {
		int len, num;

		if (cmd->cdb[2] != 0) {
			TRACE_DBG("INQUIRY: Unsupported page %x", cmd->cdb[2]);
			scst_set_cmd_error(cmd,
			    SCST_LOAD_SENSE(scst_sense_invalid_field_in_cdb));
			goto out_put;
		}

		if (virt_dev->removable)
			buf[1] = 0x80;      /* removable */
		buf[2] = 5; /* Device complies to SPC-3 */
		buf[3] = 0x02;	/* Data in format specified in SPC */
		if (cmd->tgtt->fake_aca)
			buf[3] |= 0x20;
		buf[4] = 31;/* n - 4 = 35 - 4 = 31 for full 36 byte data */
		if (scst_impl_alua_configured(cmd->dev))
			buf[5] = SCST_INQ_TPGS_MODE_IMPLICIT;
		buf[6] = 0x10; /* MultiP 1 */
		buf[7] = 2; /* CMDQUE 1, BQue 0 => commands queuing supported */

		/*
		 * 8 byte ASCII Vendor Identification of the target
		 * - left aligned.
		 */
		if (cmd->tgtt->vendor)
			memcpy(&buf[8], cmd->tgtt->vendor, 8);
		else
			memcpy(&buf[8], SCST_FIO_VENDOR, 8);

		/*
		 * 16 byte ASCII Product Identification of the target - left
		 * aligned.
		 */
		memset(&buf[16], ' ', 16);
		if (cmd->tgtt->get_product_id)
			cmd->tgtt->get_product_id(cmd->tgt_dev, &buf[16], 16);
		else {
			len = min_t(size_t, strlen(virt_dev->name), 16);
			memcpy(&buf[16], virt_dev->name, len);
		}

		/*
		 * 4 byte ASCII Product Revision Level of the target - left
		 * aligned.
		 */
		if (cmd->tgtt->revision)
			memcpy(&buf[32], cmd->tgtt->revision, 4);
		else
			memcpy(&buf[32], SCST_FIO_REV, 4);

		/** Version descriptors **/

		buf[4] += 58 - 36;
		num = 0;

		/* SAM-3 T10/1561-D revision 14 */
		buf[58 + num] = 0x0;
		buf[58 + num + 1] = 0x76;
		num += 2;

		/* Physical transport */
		if (cmd->tgtt->get_phys_transport_version != NULL) {
			uint16_t v = cmd->tgtt->get_phys_transport_version(cmd->tgt);
			if (v != 0) {
				*((__be16 *)&buf[58 + num]) = cpu_to_be16(v);
				num += 2;
			}
		}

		/* SCSI transport */
		if (cmd->tgtt->get_scsi_transport_version != NULL) {
			*((__be16 *)&buf[58 + num]) =
				cpu_to_be16(cmd->tgtt->get_scsi_transport_version(cmd->tgt));
			num += 2;
		}

		/* SPC-3 T10/1416-D revision 23 */
		buf[58 + num] = 0x3;
		buf[58 + num + 1] = 0x12;
		num += 2;

		/* Device command set */
		if (virt_dev->command_set_version != 0) {
			*((__be16 *)&buf[58 + num]) =
				cpu_to_be16(virt_dev->command_set_version);
			num += 2;
		}

		/* Vendor specific information. */
		if (cmd->tgtt->get_vend_specific) {
			/* Skip to byte 96. */
			num = 96 - 58;
			num += cmd->tgtt->get_vend_specific(cmd->tgt_dev,
						    &buf[96], INQ_BUF_SZ - 96);
		}

		buf[4] += num;
		resp_len = buf[4] + 5;
	}

	sBUG_ON(resp_len >= INQ_BUF_SZ);

	if (length > resp_len)
		length = resp_len;
	memcpy(address, buf, length);

out_put:
	scst_put_buf_full(cmd, address);
	if (length < cmd->resp_data_len)
		scst_set_resp_data_len(cmd, length);

out_free:
	kfree(buf);

out:
	TRACE_EXIT();
	return;
}

static void pcimem_exec_request_sense(struct scst_cmd *cmd)
{
	int32_t length, sl;
	uint8_t *address;
	uint8_t b[SCST_STANDARD_SENSE_LEN];

	TRACE_ENTRY();

	sl = scst_set_sense(b, sizeof(b), cmd->dev->d_sense,
		SCST_LOAD_SENSE(scst_sense_no_sense));

	length = scst_get_buf_full(cmd, &address);
	TRACE_DBG("length %d", length);
	if (length < 0) {
		PRINT_ERROR("scst_get_buf_full() failed: %d)", length);
		scst_set_cmd_error(cmd,
			SCST_LOAD_SENSE(scst_sense_hardw_error));
		goto out;
	}

	length = min(sl, length);
	memcpy(address, b, length);
	scst_set_resp_data_len(cmd, length);

	scst_put_buf_full(cmd, address);

out:
	TRACE_EXIT();
	return;
}

static int pcimem_err_recov_pg(unsigned char *p, int pcontrol,
			       struct scst_pcimem_dev *virt_dev)
{	/* Read-Write Error Recovery page for mode_sense */
	const unsigned char err_recov_pg[] = {0x1, 0xa, 0xc0, 11, 240, 0, 0, 0,
					      5, 0, 0xff, 0xff};

	memcpy(p, err_recov_pg, sizeof(err_recov_pg));
	if (1 == pcontrol)
		memset(p + 2, 0, sizeof(err_recov_pg) - 2);
	return sizeof(err_recov_pg);
}

static int pcimem_disconnect_pg(unsigned char *p, int pcontrol,
				struct scst_pcimem_dev *virt_dev)
{	/* Disconnect-Reconnect page for mode_sense */
	const unsigned char disconnect_pg[] = {0x2, 0xe, 128, 128, 0, 10, 0, 0,
					       0, 0, 0, 0, 0, 0, 0, 0};

	memcpy(p, disconnect_pg, sizeof(disconnect_pg));
	if (1 == pcontrol)
		memset(p + 2, 0, sizeof(disconnect_pg) - 2);
	return sizeof(disconnect_pg);
}

static int pcimem_rigid_geo_pg(unsigned char *p, int pcontrol,
	struct scst_pcimem_dev *virt_dev)
{
	unsigned char geo_m_pg[] = {0x04, 0x16, 0, 0, 0, DEF_HEADS, 0, 0,
				    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
				    0x3a, 0x98/* 15K RPM */, 0, 0};
	int32_t ncyl, n, rem;
	uint64_t dividend;

	memcpy(p, geo_m_pg, sizeof(geo_m_pg));
	/*
	 * Divide virt_dev->nblocks by (DEF_HEADS * DEF_SECTORS) and store
	 * the quotient in ncyl and the remainder in rem.
	 */
	dividend = virt_dev->nblocks;
	rem = do_div(dividend, DEF_HEADS * DEF_SECTORS);
	ncyl = dividend;
	if (rem != 0)
		ncyl++;
	memcpy(&n, p + 2, sizeof(u32));
	n = n | ((__force u32)cpu_to_be32(ncyl) >> 8);
	memcpy(p + 2, &n, sizeof(u32));
	if (1 == pcontrol)
		memset(p + 2, 0, sizeof(geo_m_pg) - 2);
	return sizeof(geo_m_pg);
}

static int pcimem_format_pg(unsigned char *p, int pcontrol,
			    struct scst_pcimem_dev *virt_dev)
{       /* Format device page for mode_sense */
	const unsigned char format_pg[] = {0x3, 0x16, 0, 0, 0, 0, 0, 0,
					   0, 0, 0, 0, 0, 0, 0, 0,
					   0, 0, 0, 0, 0x40, 0, 0, 0};

	memcpy(p, format_pg, sizeof(format_pg));
	p[10] = (DEF_SECTORS >> 8) & 0xff;
	p[11] = DEF_SECTORS & 0xff;
	p[12] = (virt_dev->block_size >> 8) & 0xff;
	p[13] = virt_dev->block_size & 0xff;
	if (1 == pcontrol)
		memset(p + 2, 0, sizeof(format_pg) - 2);
	return sizeof(format_pg);
}

static int pcimem_caching_pg(unsigned char *p, int pcontrol,
			     struct scst_pcimem_dev *virt_dev)
{	/* Caching page for mode_sense */
	const unsigned char caching_pg[] = {0x8, 18, 0x10, 0, 0xff, 0xff, 0, 0,
		0xff, 0xff, 0xff, 0xff, 0x80, 0x14, 0, 0, 0, 0, 0, 0};

	memcpy(p, caching_pg, sizeof(caching_pg));
	p[2] |= !(virt_dev->wt_flag || virt_dev->nv_cache) ? WCE : 0;
	if (1 == pcontrol)
		memset(p + 2, 0, sizeof(caching_pg) - 2);
	return sizeof(caching_pg);
}

static int pcimem_ctrl_m_pg(unsigned char *p, int pcontrol,
			    struct scst_pcimem_dev *virt_dev)
{	/* Control mode page for mode_sense */
	const unsigned char ctrl_m_pg[] = {0xa, 0xa, 0, 0, 0, 0, 0, 0,
					   0, 0, 0x2, 0x4b};

	memcpy(p, ctrl_m_pg, sizeof(ctrl_m_pg));
	switch (pcontrol) {
	case 0:
		p[2] |= virt_dev->dev->tst << 5;
		p[2] |= virt_dev->dev->d_sense << 2;
		p[3] |= virt_dev->dev->queue_alg << 4;
		p[4] |= virt_dev->dev->swp << 3;
		p[5] |= virt_dev->dev->tas << 6;
		break;
	case 1:
		memset(p + 2, 0, sizeof(ctrl_m_pg) - 2);
#if 0	/*
	 * It's too early to implement it, since we can't control the
	 * backstorage device parameters. ToDo
	 */
		p[2] |= 7 << 5;		/* TST */
		p[3] |= 0xF << 4;	/* QUEUE ALGORITHM MODIFIER */
#endif
		p[2] |= 1 << 2;		/* D_SENSE */
		p[4] |= 1 << 3;		/* SWP */
		p[5] |= 1 << 6;		/* TAS */
		break;
	case 2:
		p[2] |= DEF_TST << 5;
		p[2] |= DEF_DSENSE << 2;
		if (virt_dev->wt_flag || virt_dev->nv_cache)
			p[3] |= DEF_QUEUE_ALG_WT << 4;
		else
			p[3] |= DEF_QUEUE_ALG << 4;
		p[4] |= DEF_SWP << 3;
		p[5] |= DEF_TAS << 6;
		break;
	default:
		sBUG();
	}
	return sizeof(ctrl_m_pg);
}

static int pcimem_iec_m_pg(unsigned char *p, int pcontrol,
			   struct scst_pcimem_dev *virt_dev)
{	/* Informational Exceptions control mode page for mode_sense */
	const unsigned char iec_m_pg[] = {0x1c, 0xa, 0x08, 0, 0, 0, 0, 0,
					  0, 0, 0x0, 0x0};
	memcpy(p, iec_m_pg, sizeof(iec_m_pg));
	if (1 == pcontrol)
		memset(p + 2, 0, sizeof(iec_m_pg) - 2);
	return sizeof(iec_m_pg);
}

static void pcimem_exec_mode_sense(struct scst_cmd *cmd)
{
	int32_t length;
	uint8_t *address;
	uint8_t *buf;
	struct scst_pcimem_dev *virt_dev;
	uint32_t blocksize;
	uint64_t nblocks;
	unsigned char dbd, type;
	int pcontrol, pcode, subpcode;
	unsigned char dev_spec;
	int msense_6, offset = 0, len;
	unsigned char *bp;

	TRACE_ENTRY();

	buf = kzalloc(MSENSE_BUF_SZ, GFP_KERNEL);
	if (buf == NULL) {
		scst_set_busy(cmd);
		goto out;
	}

	virt_dev = cmd->dev->dh_priv;
	blocksize = virt_dev->block_size;
	nblocks = virt_dev->nblocks;

	type = cmd->dev->type;    /* type dev */
	dbd = cmd->cdb[1] & DBD;
	pcontrol = (cmd->cdb[2] & 0xc0) >> 6;
	pcode = cmd->cdb[2] & 0x3f;
	subpcode = cmd->cdb[3];
	msense_6 = (MODE_SENSE == cmd->cdb[0]);
	dev_spec = (virt_dev->dev->rd_only ||
		     cmd->tgt_dev->acg_dev->rd_only) ? WP : 0;

	dev_spec |= DPOFUA;

	length = scst_get_buf_full(cmd, &address);
	if (unlikely(length <= 0)) {
		if (length < 0) {
			PRINT_ERROR("scst_get_buf_full() failed: %d", length);
			scst_set_cmd_error(cmd,
				SCST_LOAD_SENSE(scst_sense_hardw_error));
		}
		goto out_free;
	}

	if (0x3 == pcontrol) {
		TRACE_DBG("%s", "MODE SENSE: Saving values not supported");
		scst_set_cmd_error(cmd,
		    SCST_LOAD_SENSE(scst_sense_saving_params_unsup));
		goto out_put;
	}

	if (msense_6) {
		buf[1] = type;
		buf[2] = dev_spec;
		offset = 4;
	} else {
		buf[2] = type;
		buf[3] = dev_spec;
		offset = 8;
	}

	if (0 != subpcode) {
		/* TODO: Control Extension page */
		TRACE_DBG("%s", "MODE SENSE: Only subpage 0 is supported");
		scst_set_cmd_error(cmd,
		    SCST_LOAD_SENSE(scst_sense_invalid_field_in_cdb));
		goto out_put;
	}

	if (!dbd) {
		/* Create block descriptor */
		buf[offset - 1] = 0x08;		/* block descriptor length */
		if (nblocks >> 32) {
			buf[offset + 0] = 0xFF;
			buf[offset + 1] = 0xFF;
			buf[offset + 2] = 0xFF;
			buf[offset + 3] = 0xFF;
		} else {
			/* num blks */
			buf[offset + 0] = (nblocks >> (BYTE * 3)) & 0xFF;
			buf[offset + 1] = (nblocks >> (BYTE * 2)) & 0xFF;
			buf[offset + 2] = (nblocks >> (BYTE * 1)) & 0xFF;
			buf[offset + 3] = (nblocks >> (BYTE * 0)) & 0xFF;
		}
		buf[offset + 4] = 0;			/* density code */
		buf[offset + 5] = (blocksize >> (BYTE * 2)) & 0xFF;/* blklen */
		buf[offset + 6] = (blocksize >> (BYTE * 1)) & 0xFF;
		buf[offset + 7] = (blocksize >> (BYTE * 0)) & 0xFF;

		offset += 8;			/* increment offset */
	}

	bp = buf + offset;

	switch (pcode) {
	case 0x1:	/* Read-Write error recovery page, direct access */
		len = pcimem_err_recov_pg(bp, pcontrol, virt_dev);
		break;
	case 0x2:	/* Disconnect-Reconnect page, all devices */
		len = pcimem_disconnect_pg(bp, pcontrol, virt_dev);
		break;
	case 0x3:       /* Format device page, direct access */
		len = pcimem_format_pg(bp, pcontrol, virt_dev);
		break;
	case 0x4:	/* Rigid disk geometry */
		len = pcimem_rigid_geo_pg(bp, pcontrol, virt_dev);
		break;
	case 0x8:	/* Caching page, direct access */
		len = pcimem_caching_pg(bp, pcontrol, virt_dev);
		break;
	case 0xa:	/* Control Mode page, all devices */
		len = pcimem_ctrl_m_pg(bp, pcontrol, virt_dev);
		break;
	case 0x1c:	/* Informational Exceptions Mode page, all devices */
		len = pcimem_iec_m_pg(bp, pcontrol, virt_dev);
		break;
	case 0x3f:	/* Read all Mode pages */
		len = pcimem_err_recov_pg(bp, pcontrol, virt_dev);
		len += pcimem_disconnect_pg(bp + len, pcontrol, virt_dev);
		len += pcimem_format_pg(bp + len, pcontrol, virt_dev);
		len += pcimem_caching_pg(bp + len, pcontrol, virt_dev);
		len += pcimem_ctrl_m_pg(bp + len, pcontrol, virt_dev);
		len += pcimem_iec_m_pg(bp + len, pcontrol, virt_dev);
		len += pcimem_rigid_geo_pg(bp + len, pcontrol, virt_dev);
		break;
	default:
		TRACE_DBG("MODE SENSE: Unsupported page %x", pcode);
		scst_set_cmd_error(cmd,
		    SCST_LOAD_SENSE(scst_sense_invalid_field_in_cdb));
		goto out_put;
	}

	offset += len;

	if (msense_6)
		buf[0] = offset - 1;
	else {
		buf[0] = ((offset - 2) >> 8) & 0xff;
		buf[1] = (offset - 2) & 0xff;
	}

	if (offset > length)
		offset = length;
	memcpy(address, buf, offset);

out_put:
	scst_put_buf_full(cmd, address);
	if (offset < cmd->resp_data_len)
		scst_set_resp_data_len(cmd, offset);

out_free:
	kfree(buf);

out:
	TRACE_EXIT();
	return;
}

static int pcimem_set_wt(struct scst_pcimem_dev *virt_dev, int wt)
{
	int res = 0;

	TRACE_ENTRY();

	if ((virt_dev->wt_flag == wt) || virt_dev->nv_cache)
		goto out;

	spin_lock(&virt_dev->flags_lock);
	virt_dev->wt_flag = wt;
	spin_unlock(&virt_dev->flags_lock);

	scst_dev_del_all_thr_data(virt_dev->dev);

out:
	TRACE_EXIT_RES(res);
	return res;
}

static void pcimem_ctrl_m_pg_select(unsigned char *p,
	struct scst_pcimem_dev *virt_dev, struct scst_cmd *cmd)
{
	struct scst_device *dev = virt_dev->dev;
	int old_swp = dev->swp, old_tas = dev->tas, old_dsense = dev->d_sense;

#if 0 /* Not implemented yet, see comment in vdisk_ctrl_m_pg() */
	dev->tst = (p[2] >> 5) & 1;
	dev->queue_alg = p[3] >> 4;
#else
	if ((dev->tst != ((p[2] >> 5) & 1)) || (dev->queue_alg != (p[3] >> 4))) {
		TRACE(TRACE_MINOR|TRACE_SCSI, "%s", "MODE SELECT: Changing of "
			"TST and QUEUE ALGORITHM not supported");
		scst_set_cmd_error(cmd,
		    SCST_LOAD_SENSE(scst_sense_invalid_field_in_cdb));
		return;
	}
#endif
	dev->swp = (p[4] & 0x8) >> 3;
	dev->tas = (p[5] & 0x40) >> 6;
	dev->d_sense = (p[2] & 0x4) >> 2;

	PRINT_INFO("Device %s: new control mode page parameters: SWP %x "
		"(was %x), TAS %x (was %x), D_SENSE %d (was %d)",
		virt_dev->name, dev->swp, old_swp, dev->tas, old_tas,
		dev->d_sense, old_dsense);
	return;
}

static void pcimem_exec_mode_select(struct scst_cmd *cmd)
{
	int32_t length;
	uint8_t *address;
	struct scst_pcimem_dev *virt_dev;
	int mselect_6, offset;

	TRACE_ENTRY();

	virt_dev = cmd->dev->dh_priv;
	mselect_6 = (MODE_SELECT == cmd->cdb[0]);

	length = scst_get_buf_full(cmd, &address);
	if (unlikely(length <= 0)) {
		if (length < 0) {
			PRINT_ERROR("scst_get_buf_full() failed: %d", length);
			scst_set_cmd_error(cmd,
				SCST_LOAD_SENSE(scst_sense_hardw_error));
		}
		goto out;
	}

	if (!(cmd->cdb[1] & PF) || (cmd->cdb[1] & SP)) {
		TRACE(TRACE_MINOR|TRACE_SCSI, "MODE SELECT: Unsupported "
			"value(s) of PF and/or SP bits (cdb[1]=%x)",
			cmd->cdb[1]);
		scst_set_cmd_error(cmd,
		    SCST_LOAD_SENSE(scst_sense_invalid_field_in_cdb));
		goto out_put;
	}

	if (mselect_6)
		offset = 4;
	else
		offset = 8;

	if (address[offset - 1] == 8) {
		offset += 8;
	} else if (address[offset - 1] != 0) {
		PRINT_ERROR("%s", "MODE SELECT: Wrong parameters list "
			"lenght");
		scst_set_cmd_error(cmd,
		    SCST_LOAD_SENSE(scst_sense_invalid_field_in_parm_list));
		goto out_put;
	}

	while (length > offset + 2) {
		if (address[offset] & PS) {
			PRINT_ERROR("%s", "MODE SELECT: Illegal PS bit");
			scst_set_cmd_error(cmd, SCST_LOAD_SENSE(
				scst_sense_invalid_field_in_parm_list));
			goto out_put;
		}
		if ((address[offset] & 0x3f) == 0x8) {
			/* Caching page */
			if (address[offset + 1] != 18) {
				PRINT_ERROR("%s", "MODE SELECT: Invalid "
					"caching page request");
				scst_set_cmd_error(cmd, SCST_LOAD_SENSE(
				    scst_sense_invalid_field_in_parm_list));
				goto out_put;
			}
			if (pcimem_set_wt(virt_dev,
			      (address[offset + 2] & WCE) ? 0 : 1) != 0) {
				scst_set_cmd_error(cmd,
				    SCST_LOAD_SENSE(scst_sense_hardw_error));
				goto out_put;
			}
			break;
		} else if ((address[offset] & 0x3f) == 0xA) {
			/* Control page */
			if (address[offset + 1] != 0xA) {
				PRINT_ERROR("%s", "MODE SELECT: Invalid "
					"control page request");
				scst_set_cmd_error(cmd, SCST_LOAD_SENSE(
				    scst_sense_invalid_field_in_parm_list));
				goto out_put;
			}
			pcimem_ctrl_m_pg_select(&address[offset], virt_dev, cmd);
		} else {
			PRINT_ERROR("MODE SELECT: Invalid request %x",
				address[offset] & 0x3f);
			scst_set_cmd_error(cmd, SCST_LOAD_SENSE(
			    scst_sense_invalid_field_in_parm_list));
			goto out_put;
		}
		offset += address[offset + 1];
	}

out_put:
	scst_put_buf_full(cmd, address);

out:
	TRACE_EXIT();
	return;
}

static void pcimem_exec_log(struct scst_cmd *cmd)
{
	TRACE_ENTRY();

	/* No log pages are supported */
	scst_set_cmd_error(cmd,
		SCST_LOAD_SENSE(scst_sense_invalid_field_in_cdb));

	TRACE_EXIT();
	return;
}

static void pcimem_exec_read_capacity(struct scst_cmd *cmd)
{
	int32_t length;
	uint8_t *address;
	struct scst_pcimem_dev *virt_dev;
	uint32_t blocksize;
	uint64_t nblocks;
	uint8_t buffer[8];

	TRACE_ENTRY();

	virt_dev = cmd->dev->dh_priv;
	blocksize = virt_dev->block_size;
	nblocks = virt_dev->nblocks;

	if ((cmd->cdb[8] & 1) == 0) {
		uint64_t lba = be64_to_cpu(get_unaligned((__be64 *)&cmd->cdb[2]));
		if (lba != 0) {
			TRACE_DBG("PMI zero and LBA not zero (cmd %p)", cmd);
			scst_set_cmd_error(cmd,
			    SCST_LOAD_SENSE(scst_sense_invalid_field_in_cdb));
			goto out;
		}
	}

	/* Last block on the virt_dev is (nblocks-1) */
	memset(buffer, 0, sizeof(buffer));

	/*
	 * If we are thinly provisioned, we must ensure that the initiator
	 * issues a READ_CAPACITY(16) so we can return the TPE bit. By
	 * returning 0xFFFFFFFF we do that.
	 */
	if (nblocks >> 32 || virt_dev->thin_provisioned) {
		buffer[0] = 0xFF;
		buffer[1] = 0xFF;
		buffer[2] = 0xFF;
		buffer[3] = 0xFF;
	} else {
		buffer[0] = ((nblocks - 1) >> (BYTE * 3)) & 0xFF;
		buffer[1] = ((nblocks - 1) >> (BYTE * 2)) & 0xFF;
		buffer[2] = ((nblocks - 1) >> (BYTE * 1)) & 0xFF;
		buffer[3] = ((nblocks - 1) >> (BYTE * 0)) & 0xFF;
	}
	buffer[4] = (blocksize >> (BYTE * 3)) & 0xFF;
	buffer[5] = (blocksize >> (BYTE * 2)) & 0xFF;
	buffer[6] = (blocksize >> (BYTE * 1)) & 0xFF;
	buffer[7] = (blocksize >> (BYTE * 0)) & 0xFF;

	length = scst_get_buf_full(cmd, &address);
	if (unlikely(length <= 0)) {
		if (length < 0) {
			PRINT_ERROR("scst_get_buf_full() failed: %d", length);
			scst_set_cmd_error(cmd,
				SCST_LOAD_SENSE(scst_sense_hardw_error));
		}
		goto out;
	}

	length = min_t(int, length, sizeof(buffer));

	memcpy(address, buffer, length);

	scst_put_buf_full(cmd, address);

	if (length < cmd->resp_data_len)
		scst_set_resp_data_len(cmd, length);

out:
	TRACE_EXIT();
	return;
}

static void pcimem_exec_read_capacity16(struct scst_cmd *cmd)
{
	int32_t length;
	uint8_t *address;
	struct scst_pcimem_dev *virt_dev;
	uint32_t blocksize;
	uint64_t nblocks;
	uint8_t buffer[32];

	TRACE_ENTRY();

	virt_dev = cmd->dev->dh_priv;
	blocksize = virt_dev->block_size;
	nblocks = virt_dev->nblocks - 1;

	if ((cmd->cdb[14] & 1) == 0) {
		uint64_t lba = be64_to_cpu(get_unaligned((__be64 *)&cmd->cdb[2]));
		if (lba != 0) {
			TRACE_DBG("PMI zero and LBA not zero (cmd %p)", cmd);
			scst_set_cmd_error(cmd,
			    SCST_LOAD_SENSE(scst_sense_invalid_field_in_cdb));
			goto out;
		}
	}

	memset(buffer, 0, sizeof(buffer));

	buffer[0] = nblocks >> 56;
	buffer[1] = (nblocks >> 48) & 0xFF;
	buffer[2] = (nblocks >> 40) & 0xFF;
	buffer[3] = (nblocks >> 32) & 0xFF;
	buffer[4] = (nblocks >> 24) & 0xFF;
	buffer[5] = (nblocks >> 16) & 0xFF;
	buffer[6] = (nblocks >> 8) & 0xFF;
	buffer[7] = nblocks & 0xFF;

	buffer[8] = (blocksize >> (BYTE * 3)) & 0xFF;
	buffer[9] = (blocksize >> (BYTE * 2)) & 0xFF;
	buffer[10] = (blocksize >> (BYTE * 1)) & 0xFF;
	buffer[11] = (blocksize >> (BYTE * 0)) & 0xFF;

	switch (blocksize) {
	case 512:
		buffer[13] = 3;
		break;
	case 1024:
		buffer[13] = 2;
		break;
	case 2048:
		buffer[13] = 1;
		break;
	case 4096:
	default:
		buffer[13] = 0;
		break;
	}

	if (virt_dev->thin_provisioned) {
		buffer[14] |= 0x80;     /* Add TPE */
#if 0  /*
	* Might be a big performance and functionality win, but might be
	* dangerous as well, although generally nearly always it should be set,
	* because nearly all devices should return zero for unmapped blocks.
	* But let's be on the safe side and disable it for now.
	*/
		buffer[14] |= 0x40;     /* Add TPRZ */
#endif
	}

	length = scst_get_buf_full(cmd, &address);
	if (unlikely(length <= 0)) {
		if (length < 0) {
			PRINT_ERROR("scst_get_buf_full() failed: %d", length);
			scst_set_cmd_error(cmd,
				SCST_LOAD_SENSE(scst_sense_hardw_error));
			}
		goto out;
	}

	length = min_t(int, length, sizeof(buffer));

	memcpy(address, buffer, length);

	scst_put_buf_full(cmd, address);

	if (length < cmd->resp_data_len)
		scst_set_resp_data_len(cmd, length);

out:
	TRACE_EXIT();
	return;
}

/* SPC-4 REPORT TARGET PORT GROUPS command */
static void pcimem_exec_report_tpgs(struct scst_cmd *cmd)
{
	struct scst_device *dev;
	uint8_t *address;
	void *buf;
	int32_t buf_len;
	uint32_t allocation_length, data_length, length;
	uint8_t data_format;
	int res;

	TRACE_ENTRY();

	buf_len = scst_get_buf_full(cmd, &address);
	if (buf_len < 0) {
		PRINT_ERROR("scst_get_buf_full() failed: %d", buf_len);
		scst_set_cmd_error(cmd,
			SCST_LOAD_SENSE(scst_sense_hardw_error));
		goto out;
	}

	if (cmd->cdb_len < 12)
		PRINT_WARNING("received invalid REPORT TARGET PORT GROUPS "
			      "command - length %d is too small (should be at "
			      "least 12 bytes)", cmd->cdb_len);

	dev = cmd->dev;
	data_format = cmd->cdb_len > 1 ? cmd->cdb[1] >> 5 : 0;
	allocation_length = cmd->cdb_len >= 10 ?
		be32_to_cpu(get_unaligned((__be32 *)(cmd->cdb + 6))) : 1024;

	res = scst_tg_get_group_info(&buf, &data_length, dev, data_format);
	if (res == -ENOMEM) {
		scst_set_busy(cmd);
		goto out_put;
	} else if (res < 0) {
		scst_set_cmd_error(cmd,
			SCST_LOAD_SENSE(scst_sense_invalid_field_in_cdb));
		goto out_put;
	}

	length = min_t(uint32_t, min(allocation_length, data_length), buf_len);
	memcpy(address, buf, length);
	kfree(buf);
	if (length < cmd->resp_data_len)
		scst_set_resp_data_len(cmd, length);

out_put:
	scst_put_buf_full(cmd, address);

out:
	TRACE_EXIT();
	return;
}

static void pcimem_exec_read_toc(struct scst_cmd *cmd)
{
	int32_t length, off = 0;
	uint8_t *address;
	struct scst_pcimem_dev *virt_dev;
	uint32_t nblocks;
	uint8_t buffer[4+8+8] = { 0x00, 0x0a, 0x01, 0x01, 0x00, 0x14,
				  0x01, 0x00, 0x00, 0x00, 0x00, 0x00 };

	TRACE_ENTRY();

	if (cmd->dev->type != TYPE_ROM) {
		PRINT_ERROR("%s", "READ TOC for non-CDROM device");
		scst_set_cmd_error(cmd,
			SCST_LOAD_SENSE(scst_sense_invalid_opcode));
		goto out;
	}

	if (cmd->cdb[2] & 0x0e/*Format*/) {
		PRINT_ERROR("%s", "READ TOC: invalid requested data format");
		scst_set_cmd_error(cmd,
			SCST_LOAD_SENSE(scst_sense_invalid_field_in_cdb));
		goto out;
	}

	if ((cmd->cdb[6] != 0 && (cmd->cdb[2] & 0x01)) ||
	    (cmd->cdb[6] > 1 && cmd->cdb[6] != 0xAA)) {
		PRINT_ERROR("READ TOC: invalid requested track number %x",
			cmd->cdb[6]);
		scst_set_cmd_error(cmd,
			SCST_LOAD_SENSE(scst_sense_invalid_field_in_cdb));
		goto out;
	}

	length = scst_get_buf_full(cmd, &address);
	if (unlikely(length <= 0)) {
		if (length < 0) {
			PRINT_ERROR("scst_get_buf_full() failed: %d", length);
			scst_set_cmd_error(cmd,
				SCST_LOAD_SENSE(scst_sense_hardw_error));
		}
		goto out;
	}

	virt_dev = cmd->dev->dh_priv;
	/* ToDo when you have > 8TB ROM device. */
	nblocks = (uint32_t)virt_dev->nblocks;

	/* Header */
	memset(buffer, 0, sizeof(buffer));
	buffer[2] = 0x01;    /* First Track/Session */
	buffer[3] = 0x01;    /* Last Track/Session */
	off = 4;
	if (cmd->cdb[6] <= 1) {
		/* Fistr TOC Track Descriptor */
		/* ADDR    0x10 - Q Sub-channel encodes current position data
		   CONTROL 0x04 - Data track, recoreded uninterrupted */
		buffer[off+1] = 0x14;
		/* Track Number */
		buffer[off+2] = 0x01;
		off += 8;
	}
	if (!(cmd->cdb[2] & 0x01)) {
		/* Lead-out area TOC Track Descriptor */
		buffer[off+1] = 0x14;
		/* Track Number */
		buffer[off+2] = 0xAA;
		/* Track Start Address */
		buffer[off+4] = (nblocks >> (BYTE * 3)) & 0xFF;
		buffer[off+5] = (nblocks >> (BYTE * 2)) & 0xFF;
		buffer[off+6] = (nblocks >> (BYTE * 1)) & 0xFF;
		buffer[off+7] = (nblocks >> (BYTE * 0)) & 0xFF;
		off += 8;
	}

	buffer[1] = off - 2;    /* Data  Length */

	if (off > length)
		off = length;
	memcpy(address, buffer, off);

	scst_put_buf_full(cmd, address);

	if (off < cmd->resp_data_len)
		scst_set_resp_data_len(cmd, off);

out:
	TRACE_EXIT();
	return;
}

static void pcimem_exec_prevent_allow_medium_removal(struct scst_cmd *cmd)
{
	struct scst_pcimem_dev *virt_dev = cmd->dev->dh_priv;

	TRACE_DBG("PERSIST/PREVENT 0x%02x", cmd->cdb[4]);

	spin_lock(&virt_dev->flags_lock);
	virt_dev->prevent_allow_medium_removal = cmd->cdb[4] & 0x01 ? 1 : 0;
	spin_unlock(&virt_dev->flags_lock);

	return;
}

static int pcimem_fsync(loff_t loff, loff_t len, struct scst_cmd *cmd, struct scst_device *dev)
{
	int res = 0;

	TRACE_ENTRY();

	TRACE_EXIT_RES(res);
	return res;

}

static void pcimem_exec_read(struct scst_cmd *cmd, loff_t loff)
{
	TRACE_ENTRY();

	TRACE(TRACE_DEBUG, "loff=%llx, bufflen=%x", loff, cmd->bufflen);
#ifdef USE_IPCIF
	{
		struct scst_device *dev = cmd->dev;
		struct scst_pcimem_dev *virt_dev = dev->dh_priv;
		int ret;

		if (virt_dev->obmbxid == -1) {
			PRINT_ERROR("%s", "invalid obmbxid.\n");
			goto out;
		}

		if (SCST_SENSE_VALID(cmd->sense)) {
			PRINT_ERROR("%s", "sense-code has been set.\n");
			goto out;
		}

		ret = pcimem_send_cmdreq(cmd);
		if (unlikely(ret < 0)) {
			PRINT_ERROR("pcimem_send_cmdreq() error(%d).\n", ret);
			goto out;
		}

		do {
			ret = ipcscst_recv(virt_dev->obmbxid, cmd, &ipcscst_funcs);
			if (unlikely(ret < 0)) {
				PRINT_ERROR("ipcscst_recv() error(%d).\n", ret);
				goto out;
			}
		} while (virt_dev->ipc_active_cmd == cmd);

	}
out:
#endif /* USE_IPCIF */

	TRACE_EXIT();
	return;
}

static void pcimem_exec_write(struct scst_cmd *cmd, loff_t loff)
{
	TRACE_ENTRY();

	TRACE(TRACE_DEBUG, "loff=%llx, bufflen=%x", loff, cmd->bufflen);
#ifdef USE_IPCIF
	{
		struct scst_device *dev = cmd->dev;
		struct scst_pcimem_dev *virt_dev = dev->dh_priv;
		struct ipcscst_areainfo areainfo[MAX_SGV_CNT];
		int ret;
		int i;

		if (virt_dev->obmbxid == -1) {
			PRINT_ERROR("%s", "invalid obmbxid.\n");
			goto out;
		}

		if (SCST_SENSE_VALID(cmd->sense)) {
			PRINT_ERROR("%s", "sense-code has been set.\n");
			goto out;
		}

		for (i = 0; i < cmd->dh_sg_cnt; i++) {
			areainfo[i].addr = cmd->dh_sg[i].dma_address;
			areainfo[i].size = cmd->dh_sg[i].dma_length;
		}

		ret = ipcscst_send_dmadone(virt_dev->ibmbxid, cmd->dh_sn, areainfo, cmd->dh_sg_cnt, IPCSCST_DATA_WRITE);
		if (ret < 0) {
			PRINT_ERROR("ipcscst_send_dmadone() error(%d).\n", ret);
			goto out;
		}

		ret = ipcscst_recv(virt_dev->obmbxid, cmd, &ipcscst_funcs);
		if (ret < 0) {
			PRINT_ERROR("ipcscst_recv() error(%d).\n", ret);
			goto out;
		}

	}

out:
#endif

	TRACE_EXIT();
	return;
}

static void pcimem_exec_verify(struct scst_cmd *cmd, loff_t loff)
{
	TRACE_ENTRY();

	TRACE_EXIT();
	return;
}

static void pcimem_exec_write_buffer(struct scst_cmd *cmd)
{
	TRACE_ENTRY();

	TRACE(TRACE_DEBUG, "bufflen=%x", cmd->bufflen);
#ifdef USE_IPCIF
	{
		struct scst_device *dev = cmd->dev;
		struct scst_pcimem_dev *virt_dev = dev->dh_priv;
		struct ipcscst_areainfo areainfo[MAX_SGV_CNT];
		int ret;
		int i;

		if (virt_dev->obmbxid == -1) {
			PRINT_ERROR("%s", "invalid obmbxid.\n");
			goto out;
		}

		if (SCST_SENSE_VALID(cmd->sense)) {
			PRINT_ERROR("%s", "sense-code has been set.\n");
			goto out;
		}

		for (i = 0; i < cmd->dh_sg_cnt; i++) {
			areainfo[i].addr = cmd->dh_sg[i].dma_address;
			areainfo[i].size = cmd->dh_sg[i].dma_length;
		}

		ret = ipcscst_send_dmadone(virt_dev->ibmbxid, cmd->sn, areainfo, cmd->dh_sg_cnt, IPCSCST_DATA_WRITE);
		if (ret < 0) {
			PRINT_ERROR("ipcscst_send_dmadone() error(%d).\n", ret);
			goto out;
		}

		ret = ipcscst_recv(virt_dev->obmbxid, cmd, &ipcscst_funcs);
		if (ret < 0) {
			PRINT_ERROR("ipcscst_recv() error(%d).\n", ret);
			goto out;
		}

	}

out:
#endif

	TRACE_EXIT();
	return;
}

int pcimem_do_job(struct scst_cmd *cmd)
{
	int rc, res;
	uint64_t lba_start = 0;
	loff_t data_len = 0;
	uint8_t *cdb = cmd->cdb;
	int opcode = cdb[0];
	loff_t loff;
	struct scst_device *dev = cmd->dev;
	struct scst_pcimem_dev *virt_dev = dev->dh_priv;
	int fua = 0;

	TRACE_ENTRY();

	TRACE(TRACE_DEBUG, "opcode=0x%02x\n", opcode);

	if (!virt_dev->media_exists && (opcode != INQUIRY)) {
		TRACE_DBG("%s", "media not exists");
		scst_set_cmd_error(cmd,
			SCST_LOAD_SENSE(scst_sense_not_ready));
		goto out_done;
	}

	if (virt_dev->media_changed && scst_is_ua_command(cmd)) {
		spin_lock(&virt_dev->flags_lock);
		if (virt_dev->media_changed) {
			virt_dev->media_changed = 0;
			TRACE_DBG("%s", "Reporting media changed");
			scst_set_cmd_error(cmd,
				SCST_LOAD_SENSE(scst_sense_medium_changed_UA));
			spin_unlock(&virt_dev->flags_lock);
			goto out_done;
		}
		spin_unlock(&virt_dev->flags_lock);
	}

	switch (cmd->queue_type) {
	case SCST_CMD_QUEUE_ORDERED:
		TRACE(TRACE_ORDER, "ORDERED cmd %p (op %x)", cmd, cmd->cdb[0]);
		break;
	case SCST_CMD_QUEUE_HEAD_OF_QUEUE:
		TRACE(TRACE_ORDER, "HQ cmd %p (op %x)", cmd, cmd->cdb[0]);
		break;
	default:
		break;
	}

	rc = scst_check_local_events(cmd);
	if (unlikely(rc != 0))
		goto out_done;

	cmd->status = 0;
	cmd->msg_status = 0;
	cmd->host_status = DID_OK;
	cmd->driver_status = 0;

	switch (opcode) {
	case READ_6:
	case WRITE_6:
	case VERIFY_6:
		lba_start = (((cdb[1] & 0x1f) << (BYTE * 2)) +
			     (cdb[2] << (BYTE * 1)) +
			     (cdb[3] << (BYTE * 0)));
		data_len = cmd->bufflen;
		break;
	case READ_10:
	case READ_12:
	case WRITE_10:
	case WRITE_12:
	case VERIFY:
	case WRITE_VERIFY:
	case WRITE_VERIFY_12:
	case VERIFY_12:
		lba_start |= ((u64)cdb[2]) << 24;
		lba_start |= ((u64)cdb[3]) << 16;
		lba_start |= ((u64)cdb[4]) << 8;
		lba_start |= ((u64)cdb[5]);
		data_len = cmd->bufflen;
		break;
	case READ_16:
	case WRITE_16:
	case WRITE_VERIFY_16:
	case VERIFY_16:
		lba_start |= ((u64)cdb[2]) << 56;
		lba_start |= ((u64)cdb[3]) << 48;
		lba_start |= ((u64)cdb[4]) << 40;
		lba_start |= ((u64)cdb[5]) << 32;
		lba_start |= ((u64)cdb[6]) << 24;
		lba_start |= ((u64)cdb[7]) << 16;
		lba_start |= ((u64)cdb[8]) << 8;
		lba_start |= ((u64)cdb[9]);
		data_len = cmd->bufflen;
		break;
	case SYNCHRONIZE_CACHE:
		lba_start |= ((u64)cdb[2]) << 24;
		lba_start |= ((u64)cdb[3]) << 16;
		lba_start |= ((u64)cdb[4]) << 8;
		lba_start |= ((u64)cdb[5]);
		data_len = ((cdb[7] << (BYTE * 1)) + (cdb[8] << (BYTE * 0)))
				<< virt_dev->block_shift;
		if (data_len == 0)
			data_len = virt_dev->file_size -
				((loff_t)lba_start << virt_dev->block_shift);
		break;
	}

	loff = (loff_t)lba_start << virt_dev->block_shift;
	TRACE_DBG("cmd %p, lba_start %lld, loff %lld, data_len %lld", cmd,
		  (long long unsigned int)lba_start,
		  (long long unsigned int)loff,
		  (long long unsigned int)data_len);
	if (unlikely(loff < 0) || unlikely(data_len < 0) ||
	    unlikely((loff + data_len) > virt_dev->file_size)) {
		PRINT_INFO("Access beyond the end of the device "
			"(%lld of %lld, len %lld)",
			   (long long unsigned int)loff,
			   (long long unsigned int)virt_dev->file_size,
			   (long long unsigned int)data_len);
		scst_set_cmd_error(cmd, SCST_LOAD_SENSE(
					scst_sense_block_out_range_error));
		goto out_compl;
	}

	switch (opcode) {
	case WRITE_10:
	case WRITE_12:
	case WRITE_16:
		fua = (cdb[1] & 0x8);
		if (fua) {
			TRACE(TRACE_ORDER, "FUA: loff=%lld, "
				"data_len=%lld", (long long unsigned int)loff,
				(long long unsigned int)data_len);
		}
		break;
	}

	switch (opcode) {
	case READ_6:
	case READ_10:
	case READ_12:
	case READ_16:
		pcimem_exec_read(cmd, loff);
		break;
	case WRITE_6:
	case WRITE_10:
	case WRITE_12:
	case WRITE_16:
	{
		pcimem_exec_write(cmd, loff);
		/* O_SYNC flag is used for WT devices */
		if (fua)
			pcimem_fsync(loff, data_len, cmd, dev);
		break;
	}
	case WRITE_VERIFY:
	case WRITE_VERIFY_12:
	case WRITE_VERIFY_16:
	{
		/* ToDo: BLOCKIO VERIFY */
		pcimem_exec_write(cmd, loff);
		/* O_SYNC flag is used for WT devices */
		if (scsi_status_is_good(cmd->status))
			pcimem_exec_verify(cmd, loff);
		break;
	}
	case SYNCHRONIZE_CACHE:
	{
		int immed = cdb[1] & 0x2;
		TRACE(TRACE_ORDER, "SYNCHRONIZE_CACHE: "
			"loff=%lld, data_len=%lld, immed=%d",
			(long long unsigned int)loff,
			(long long unsigned int)data_len, immed);
		if (immed) {
			scst_cmd_get(cmd); /* to protect dev */
			cmd->completed = 1;
			cmd->scst_cmd_done(cmd, SCST_CMD_STATE_DEFAULT,
				SCST_CONTEXT_SAME);
			pcimem_fsync(loff, data_len, NULL, dev);
			/* ToDo: pcimem_fsync() error processing */
			scst_cmd_put(cmd);
			goto out_thr;
		} else {
			pcimem_fsync(loff, data_len, cmd, dev);
			break;
		}
	}
	case VERIFY_6:
	case VERIFY:
	case VERIFY_12:
	case VERIFY_16:
		pcimem_exec_verify(cmd, loff);
		break;
	case MODE_SENSE:
	case MODE_SENSE_10:
		pcimem_exec_mode_sense(cmd);
		break;
	case MODE_SELECT:
	case MODE_SELECT_10:
		pcimem_exec_mode_select(cmd);
		break;
	case LOG_SELECT:
	case LOG_SENSE:
		pcimem_exec_log(cmd);
		break;
	case ALLOW_MEDIUM_REMOVAL:
		pcimem_exec_prevent_allow_medium_removal(cmd);
		break;
	case READ_TOC:
		pcimem_exec_read_toc(cmd);
		break;
	case START_STOP:
		pcimem_fsync(0, virt_dev->file_size, cmd, dev);
		break;
	case RESERVE:
	case RESERVE_10:
	case RELEASE:
	case RELEASE_10:
	case TEST_UNIT_READY:
		break;
	case INQUIRY:
		pcimem_exec_inquiry(cmd);
		break;
	case REQUEST_SENSE:
		pcimem_exec_request_sense(cmd);
		break;
	case READ_CAPACITY:
		pcimem_exec_read_capacity(cmd);
		break;
	case SERVICE_ACTION_IN:
		if ((cmd->cdb[1] & 0x1f) == SAI_READ_CAPACITY_16) {
			pcimem_exec_read_capacity16(cmd);
			break;
		}
		goto out_invalid_opcode;
	case UNMAP:
		pcimem_exec_unmap(cmd);
		break;
	case MAINTENANCE_IN:
		switch (cmd->cdb[1] & 0x1f) {
		case MI_REPORT_TARGET_PGS:
			pcimem_exec_report_tpgs(cmd);
			break;
		default:
			goto out_invalid_opcode;
		}
		break;
	case WRITE_BUFFER:
		pcimem_exec_write_buffer(cmd);
		break;
	case REPORT_LUNS:
	default:
		goto out_invalid_opcode;
	}

out_compl:
	cmd->completed = 1;

out_done:
	cmd->scst_cmd_done(cmd, SCST_CMD_STATE_DEFAULT, SCST_CONTEXT_SAME);

out_thr:
	res = SCST_EXEC_COMPLETED;

	TRACE_EXIT_RES(res);
	return res;

out_invalid_opcode:
	TRACE_DBG("Invalid opcode %d", opcode);
	scst_set_cmd_error(cmd,
		    SCST_LOAD_SENSE(scst_sense_invalid_opcode));
	goto out_compl;
}


int scst_pcimem_allocator_init(struct scst_pcimem_dev *virt_dev)
{
	int res = 0;

	return res;
}

#ifdef USE_IPCIF
static void pcimem_proc_cmdresp(void *priv, unsigned int tag, unsigned int resp_code, unsigned int sc, unsigned int asc, unsigned int ascq, unsigned int add_inf, unsigned int add_spec)
{
	struct scst_cmd *cmd = priv;
	struct scst_device *dev = cmd->dev;
	struct scst_pcimem_dev *virt_dev = dev->dh_priv;

	//	PRINT_INFO("up(ibmbxid=%d)\n", virt_dev->ibmbxid);

	virt_dev->ipc_active_cmd = NULL;

	up(&virt_dev->mbx_sem);

	TRACE(TRACE_SCSI, "sc=0x%02x,asc=0x%02x,ascq=0x%02x\n", sc, asc, ascq);

	if (sc) {
		scst_set_cmd_error(cmd, sc, asc, ascq);
	}

	return;
}

static void pcimem_proc_dmardy(void *priv, unsigned int tag, struct ipcscst_areainfo *area, int num_area, int direction)
{
	struct scst_cmd *cmd = priv;
	//	struct scst_device *dev = cmd->dev;
	//	struct scst_pcimem_dev *virt_dev = dev->dh_priv;
	//	struct sgv_pool_obj *obj = virt_dev->obj;
	int size = 0;
	int n;
	int i, j;

	TRACE_ENTRY();

	if ((num_area <= 0) || (num_area > MAX_SGV_CNT)) {
		PRINT_ERROR("invalid sg_cnt(=%d)\n", num_area);
		cmd->sg_cnt = 0;
		return;
	}

	for (i = 0; i < num_area; i++) {
		cmd->dh_sg[cmd->dh_sg_cnt].dma_address = area[i].addr;
		cmd->dh_sg[cmd->dh_sg_cnt].dma_length = area[i].size;

		n = pcimem_add_sg_entries(&cmd->sg[cmd->sg_cnt], MAX_SGV_CNT - cmd->sg_cnt, area[i].addr, area[i].size);
		if (n <= 0) {
			PRINT_ERROR("pcimem_add_sg_entries(offset=0x%08lx,size=0x%08lx) error", area[i].addr, area[i].size);
			goto out;
		}


		for (j = 0; j < n; j++) {
			TRACE(TRACE_DEBUG, "dma_addr=0x%08x,len=0x%08x", cmd->sg[cmd->sg_cnt + j].dma_address, cmd->sg[cmd->sg_cnt + j].dma_length);
		}

		size += area[i].size;
		cmd->sg_cnt += n;

		cmd->dh_sg_cnt++;
	}

	if (size != cmd->bufflen) {
		TRACE(TRACE_DEBUG, "size mismatch(reqsize=0x%x,alloced=0x%x)", cmd->bufflen, size);

		if (cmd->data_direction & SCST_DATA_READ) {
			scst_set_resp_data_len(cmd, size);
		}
	}

out:
	TRACE_EXIT();
	return;
}

static void pcimem_proc_reset_resp(void *priv)
{
	struct scst_cmd *cmd = priv;
	struct scst_device *dev = cmd->dev;
	struct scst_pcimem_dev *virt_dev = dev->dh_priv;

	TRACE_ENTRY();

	virt_dev->ipc_active_cmd = NULL;

	up(&virt_dev->mbx_sem);

	TRACE_EXIT();
	return;
}

static int pcimem_send_cmdreq(struct scst_cmd *cmd)
{
	struct scst_device *dev = cmd->dev;
	struct scst_pcimem_dev *virt_dev = dev->dh_priv;
	int res = 0;
	int ret;

	TRACE(TRACE_SCSI, "opcode=0x%02x\n", cmd->cdb[0]);

	if (virt_dev->ibmbxid == -1) {
		PRINT_ERROR("%s", "invalid ibmbxid.\n");
		return -1;
	}

	//	PRINT_INFO("down(ibmbxid=%d)\n", virt_dev->ibmbxid);

	down(&virt_dev->mbx_sem);

	//	PRINT_INFO("down(ibmbxid=%d) OK\n", virt_dev->ibmbxid);

	virt_dev->ipc_active_cmd = cmd;
	cmd->dh_sn = virt_dev->curr_sn++;

	ret = ipcscst_send_cmdreq(virt_dev->ibmbxid, cmd->dh_sn, cmd->cdb);
	if (ret < 0) {
		PRINT_ERROR("ipcscst_send_cmdreq() error(%d).\n", ret);
		goto out_fail;
	}

out:
	return res;

out_fail:
	up(&virt_dev->mbx_sem);
	res = -1;
	goto out;
}

#if 0
static int pcimem_req_buf(struct scst_cmd *cmd, struct sgv_pool_obj *obj)
{
	struct scst_device *dev = cmd->dev;
	struct scst_pcimem_dev *virt_dev = dev->dh_priv;
	int res = 0;
	int ret;

	if (virt_dev->obmbxid == -1) {
		PRINT_ERROR("%s", "invalid obmbxid.\n");
		return -1;
	}

	ret = pcimem_send_cmdreq(cmd);
	if (ret < 0) {
		PRINT_ERROR("pcimem_send_cmdreq() error(%d).\n", ret);
		goto out;
	}

	virt_dev->obj = obj;
	virt_dev->sg_cnt = 0;
	ret = ipcscst_recv(virt_dev->obmbxid, cmd, &ipcscst_funcs);
	if (ret < 0) {
		PRINT_ERROR("ipcscst_recv() error(%d).\n", ret);
		goto out;
	}

out:
	return res;
}
#endif
#endif /* USE_IPCIF */

static bool scst_on_sg_tablesize_low(struct scst_cmd *cmd, bool out)
{
	bool res;
	int sg_cnt = out ? cmd->out_sg_cnt : cmd->sg_cnt;
	static int ll;
	struct scst_tgt_dev *tgt_dev = cmd->tgt_dev;

	TRACE_ENTRY();

	if (sg_cnt > cmd->tgt->sg_tablesize) {
		/* It's the target's side business */
		goto failed;
	}

	if (tgt_dev->dev->handler->on_sg_tablesize_low == NULL)
		goto failed;

	res = tgt_dev->dev->handler->on_sg_tablesize_low(cmd);

	TRACE_DBG("on_sg_tablesize_low(%p) returned %d", cmd, res);

out:
	TRACE_EXIT_RES(res);
	return res;

failed:
	res = false;
	if ((ll < 10) || TRACING_MINOR()) {
		PRINT_INFO("Unable to complete command due to SG IO count "
			"limitation (%srequested %d, available %d, tgt lim %d)",
			out ? "OUT buffer, " : "", cmd->sg_cnt,
			tgt_dev->max_sg_cnt, cmd->tgt->sg_tablesize);
		ll++;
	}
	goto out;
}

static struct sgv_pool_obj *pcimem_create_sg_obj(gfp_t gfp_mask, struct scst_cmd *cmd)
{
	struct sgv_pool_obj *obj;
	int sz;

	sz = sizeof(*obj) + MAX_SGV_CNT * sizeof(obj->sg_entries[0]);

	obj = kmalloc(sz, gfp_mask);
	if (unlikely(obj == NULL)) {
		TRACE(TRACE_OUT_OF_MEM, "Allocation of "
			  "sgv_pool_obj failed (size %d)", sz);
		return NULL;
	}

	memset(obj, 0, sizeof(*obj));

	obj->owner_pool = NULL;
	obj->cache_num = -1;
	obj->allocator_priv = cmd;

	obj->sg_entries = obj->sg_entries_data;
	sg_init_table(obj->sg_entries, MAX_SGV_CNT);

	return obj;
}

#if 0
static struct scatterlist *pcimem_alloc_sg_entries(unsigned int size,
												   gfp_t gfp_mask,
												   int *count,
												   struct sgv_pool_obj **sgv,
												   struct scst_cmd *cmd)
{
	struct sgv_pool_obj *obj;
	int pages, cnt, sz;
	struct scatterlist *res = NULL;
	struct scst_device *dev = cmd->dev;
	struct scst_pcimem_dev *virt_dev = dev->dh_priv;
	uint64_t lba_start = 0;
	loff_t data_len = 0;
	uint8_t *cdb = cmd->cdb;
	int opcode = cdb[0];

	TRACE_ENTRY();

	if (unlikely(size == 0))
		goto out;

	EXTRACHECKS_BUG_ON((gfp_mask & __GFP_NOFAIL) == __GFP_NOFAIL);

	switch (opcode) {
	case READ_6:
	case WRITE_6:
	case VERIFY_6:
		lba_start = (((cdb[1] & 0x1f) << (BYTE * 2)) +
			     (cdb[2] << (BYTE * 1)) +
			     (cdb[3] << (BYTE * 0)));
		data_len = size;
		break;
	case READ_10:
	case READ_12:
	case WRITE_10:
	case WRITE_12:
	case VERIFY:
	case WRITE_VERIFY:
	case WRITE_VERIFY_12:
	case VERIFY_12:
		lba_start |= ((u64)cdb[2]) << 24;
		lba_start |= ((u64)cdb[3]) << 16;
		lba_start |= ((u64)cdb[4]) << 8;
		lba_start |= ((u64)cdb[5]);
		data_len = size;
		break;
	case READ_16:
	case WRITE_16:
	case WRITE_VERIFY_16:
	case VERIFY_16:
		lba_start |= ((u64)cdb[2]) << 56;
		lba_start |= ((u64)cdb[3]) << 48;
		lba_start |= ((u64)cdb[4]) << 40;
		lba_start |= ((u64)cdb[5]) << 32;
		lba_start |= ((u64)cdb[6]) << 24;
		lba_start |= ((u64)cdb[7]) << 16;
		lba_start |= ((u64)cdb[8]) << 8;
		lba_start |= ((u64)cdb[9]);
		data_len = size;
		break;
	case SYNCHRONIZE_CACHE:
		lba_start |= ((u64)cdb[2]) << 24;
		lba_start |= ((u64)cdb[3]) << 16;
		lba_start |= ((u64)cdb[4]) << 8;
		lba_start |= ((u64)cdb[5]);
		data_len = ((cdb[7] << (BYTE * 1)) + (cdb[8] << (BYTE * 0)))
				<< virt_dev->block_shift;
		if (data_len == 0)
			data_len = virt_dev->file_size -
				((loff_t)lba_start << virt_dev->block_shift);
		break;
	}

	pages = ((size + PAGE_SIZE - 1) >> PAGE_SHIFT);
	sz = sizeof(*obj) + MAX_SGV_CNT * sizeof(obj->sg_entries[0]);

	obj = kmalloc(sz, gfp_mask);
	if (unlikely(obj == NULL)) {
		TRACE(TRACE_OUT_OF_MEM, "Allocation of "
			  "sgv_pool_obj failed (size %d)", size);
		goto out_fail;
	}

	memset(obj, 0, sizeof(*obj));

	obj->owner_pool = NULL;
	obj->cache_num = -1;
	obj->pages = pages;
	obj->allocator_priv = cmd;

	obj->sg_entries = obj->sg_entries_data;
	sg_init_table(obj->sg_entries, MAX_SGV_CNT);

#ifdef USE_IPCIF
	if (gfp_mask & GFP_ATOMIC) {
		PRINT_ERROR("%s", "scst_ipcif doesn't exec(atomic context)\n");
	}

	if ((virt_dev->ibmbxid != -1) && (virt_dev->obmbxid != -1)) {
		pcimem_req_buf(cmd, obj);
		cnt = virt_dev->sg_cnt;
	} else {
#endif
		cnt = pcimem_add_sg_entries(&obj->sg_entries[0],
									MAX_SGV_CNT,
									((loff_t)lba_start << virt_dev->block_shift,
									 PAGE_SIZE * pages);
#ifdef USE_IPCIF
	}
#endif /* USE_IPCIF */

	sg_mark_end(&obj->sg_entries[cnt]);

	*count = cnt;
	res = obj->sg_entries;
	*sgv = obj;

out:
	TRACE_EXIT_HRES(res);
	return res;

out_fail:
	res = NULL;
	*count = 0;
	*sgv = NULL;
	TRACE_MEM("%s", "Allocation failed");
	goto out;
}
#endif

static void pcimem_free_sg_entries(struct sgv_pool_obj *obj)
{
	kfree(obj);

	return;
}

static int pcimem_create_sg_objs(struct scst_cmd *cmd)
{
	struct scst_tgt_dev *tgt_dev = cmd->tgt_dev;
	struct sgv_pool_obj *obj;
	gfp_t gfp_mask;
	int atomic = scst_cmd_atomic(cmd);
	int res = -ENOMEM;

	TRACE_ENTRY();

	gfp_mask = tgt_dev->gfp_mask | (atomic ? GFP_ATOMIC : GFP_KERNEL);

	obj = pcimem_create_sg_obj(gfp_mask, cmd);
	if (!obj) {
		goto out;
	}

	cmd->dh_sg_cnt = 0;
	cmd->dh_sgv = obj;
	cmd->dh_sg = obj->sg_entries;

	obj = pcimem_create_sg_obj(gfp_mask, cmd);
	if (!obj) {
		goto out_dh_sg_free;
	}

	cmd->sg_cnt = 0;
	cmd->sgv = obj;
	cmd->sg = obj->sg_entries;

	if (unlikely(cmd->sg_cnt > tgt_dev->max_sg_cnt))
		goto out_sg_free;

	if (cmd->data_direction != SCST_DATA_BIDI)
		goto success;

	obj = pcimem_create_sg_obj(gfp_mask, cmd);
	if (!obj) {
		goto out_sg_free;
	}

	cmd->out_sg_cnt = 0;
	cmd->out_sgv = obj;
	cmd->out_sg = obj->sg_entries;

	if (unlikely(cmd->out_sg_cnt > tgt_dev->max_sg_cnt))
		goto out_out_sg_free;

success:
	cmd->no_local_mem = 1;
	cmd->dh_data_buf_alloced = 1;
	res = 0;

out:
	TRACE_EXIT_RES(res);
	return res;

out_out_sg_free:
	pcimem_free_sg_entries(cmd->out_sgv);
	cmd->out_sgv = NULL;
	cmd->out_sg = NULL;
	cmd->out_sg_cnt = 0;

out_sg_free:
	pcimem_free_sg_entries(cmd->sgv);
	cmd->sgv = NULL;
	cmd->sg = NULL;
	cmd->sg_cnt = 0;

out_dh_sg_free:
	pcimem_free_sg_entries(cmd->dh_sgv);
	cmd->dh_sgv = NULL;
	cmd->dh_sg = NULL;
	cmd->dh_sg_cnt = 0;

	goto out;
}

#if 0
static int pcimem_alloc_space(struct scst_cmd *cmd)
{
	gfp_t gfp_mask;
	int res = -ENOMEM;
	int atomic = scst_cmd_atomic(cmd);
	struct scst_tgt_dev *tgt_dev = cmd->tgt_dev;

	TRACE_ENTRY();

	gfp_mask = tgt_dev->gfp_mask | (atomic ? GFP_ATOMIC : GFP_KERNEL);

	cmd->dh_sg = pcimem_alloc_sg_entries(cmd->bufflen,
										 gfp_mask,
										 &cmd->dh_sg_cnt,
										 &cmd->dh_sgv,
										 cmd);
	if (cmd->dh_sg == NULL) {
		goto out;
	}

	cmd->sg = pcimem_alloc_sg_entries(cmd->bufflen,
									  gfp_mask,
									  &cmd->sg_cnt,
									  &cmd->sgv,
									  cmd);
	if (cmd->sg == NULL) {
		goto out_dh_sg_free;
	}

	if (unlikely(cmd->sg_cnt > tgt_dev->max_sg_cnt))
		if (!scst_on_sg_tablesize_low(cmd, false))
			goto out_sg_free;

	if (cmd->data_direction != SCST_DATA_BIDI)
		goto success;

	cmd->out_sg = pcimem_alloc_sg_entries(cmd->out_bufflen,
										  gfp_mask,
										  &cmd->out_sg_cnt,
										  &cmd->out_sgv,
										  cmd);
	if (cmd->out_sg == NULL) {
		goto out_sg_free;
	}

	if (unlikely(cmd->out_sg_cnt > tgt_dev->max_sg_cnt))
		if (!scst_on_sg_tablesize_low(cmd, true))
			goto out_out_sg_free;

success:
	cmd->no_local_mem = 1;
	cmd->dh_data_buf_alloced = 1;
	res = 0;

out:
	TRACE_EXIT();
	return res;

out_out_sg_free:
	pcimem_free_sg_entries(cmd->out_sgv);
	cmd->out_sgv = NULL;
	cmd->out_sg = NULL;
	cmd->out_sg_cnt = 0;

out_sg_free:
	pcimem_free_sg_entries(cmd->sgv);
	cmd->sgv = NULL;
	cmd->sg = NULL;
	cmd->sg_cnt = 0;

out_dh_sg_free:
	pcimem_free_sg_entries(cmd->dh_sgv);
	cmd->dh_sgv = NULL;
	cmd->dh_sg = NULL;
	cmd->dh_sg_cnt = 0;
	goto out;
}
#endif

int pcimem_alloc_data_buf(struct scst_cmd *cmd)
{
	int res = SCST_CMD_STATE_DEFAULT;

	TRACE_ENTRY();

	TRACE_MEM("opcode=0x%02x\n", cmd->cdb[0]);

	switch (cmd->cdb[0]) {
	case READ_6:
	case READ_10:
	case READ_12:
	case READ_16:
		pcimem_create_sg_objs(cmd);
		break;
	case WRITE_6:
	case WRITE_10:
	case WRITE_12:
	case WRITE_16:
		pcimem_create_sg_objs(cmd);
		break;
	case WRITE_BUFFER:
		pcimem_create_sg_objs(cmd);
		break;
	default:
		/* do nothing */
		break;
	}

#ifdef USE_IPCIF
	if (cmd->no_local_mem && cmd->data_direction & SCST_DATA_WRITE) {
		struct scst_device *dev = cmd->dev;
		struct scst_pcimem_dev *virt_dev = dev->dh_priv;
		int ret;

		ret = pcimem_send_cmdreq(cmd);
		if (unlikely(ret < 0)) {
			PRINT_ERROR("pcimem_send_cmdreq() error(%d).\n", ret);
			res = SCST_CMD_STATE_STOP;
			goto out;
		}

		ret = ipcscst_recv(virt_dev->obmbxid, cmd, &ipcscst_funcs);
		if (unlikely(ret < 0)) {
			PRINT_ERROR("ipcscst_recv() error(%d).\n", ret);
			res = SCST_CMD_STATE_STOP;
			goto out;
		}
	}
#endif /* USE_IPCIF */

	if ((!cmd->sg_cnt) && (cmd->data_direction & SCST_DATA_WRITE)) {
		/*
		 *  if allocation failed, skip transfer...
		 */
		res = SCST_CMD_STATE_PRE_DEV_DONE;
	}

out:
	TRACE_EXIT_RES(res);
	return res;
}

void pcimem_on_free_cmd(struct scst_cmd *cmd)
{
	TRACE_ENTRY();

	TRACE_MEM("opcode=0x%02x\n", cmd->cdb[0]);

	if (!cmd->dh_data_buf_alloced) {
		goto out;
	}

	if (cmd->out_sgv != NULL) {
		pcimem_free_sg_entries(cmd->out_sgv);
		cmd->out_sgv = NULL;
		cmd->out_sg_cnt = 0;
		cmd->out_sg = NULL;
		cmd->out_bufflen = 0;
	}

	if ((cmd->sg_cnt) && (cmd->data_direction & SCST_DATA_READ)) {
#ifdef USE_IPCIF
		struct scst_device *dev = cmd->dev;
		struct scst_pcimem_dev *virt_dev = dev->dh_priv;
		struct ipcscst_areainfo areainfo[MAX_SGV_CNT];
		int ret;
		int i;

		for (i = 0; i < cmd->dh_sg_cnt; i++) {
			areainfo[i].addr = cmd->dh_sg[i].dma_address;
			areainfo[i].size = cmd->dh_sg[i].dma_length;
		}

		ret = ipcscst_send_dmadone(virt_dev->ibmbxid, cmd->dh_sn, areainfo, cmd->dh_sg_cnt, IPCSCST_DATA_READ);
		if (ret < 0) {
			PRINT_ERROR("ipcscst_send_dmadone() error(%d).\n", ret);
		}
#endif /* USE_IPCIF */
	}

	pcimem_free_sg_entries(cmd->dh_sgv);
	cmd->dh_sgv = NULL;
	cmd->dh_sg_cnt = 0;
	cmd->dh_sg = NULL;

	pcimem_free_sg_entries(cmd->sgv);
	cmd->sgv = NULL;
	cmd->sg_cnt = 0;
	cmd->sg = NULL;
	cmd->bufflen = 0;
	cmd->data_len = 0;

out:
	TRACE_EXIT();
	return;
}

int pcimem_task_mgmt_fn(struct scst_mgmt_cmd *mcmd,
	struct scst_tgt_dev *tgt_dev)
{
	TRACE_ENTRY();

	if ((mcmd->fn == SCST_LUN_RESET) || (mcmd->fn == SCST_TARGET_RESET)) {
		/* Restore default values */
		struct scst_device *dev = tgt_dev->dev;
		struct scst_pcimem_dev *virt_dev = dev->dh_priv;

		dev->tst = DEF_TST;
		dev->d_sense = DEF_DSENSE;
		if (virt_dev->wt_flag && !virt_dev->nv_cache)
			dev->queue_alg = DEF_QUEUE_ALG_WT;
		else
			dev->queue_alg = DEF_QUEUE_ALG;
		dev->swp = DEF_SWP;
		dev->tas = DEF_TAS;

		spin_lock(&virt_dev->flags_lock);
		virt_dev->prevent_allow_medium_removal = 0;
		spin_unlock(&virt_dev->flags_lock);
	} else if (mcmd->fn == SCST_PR_ABORT_ALL) {
		struct scst_device *dev = tgt_dev->dev;
		struct scst_pcimem_dev *virt_dev = dev->dh_priv;
		spin_lock(&virt_dev->flags_lock);
		virt_dev->prevent_allow_medium_removal = 0;
		spin_unlock(&virt_dev->flags_lock);
	}

	TRACE_EXIT();
	return SCST_DEV_TM_NOT_COMPLETED;
}

int pcimem_add_virtual_device(struct scst_pcimem_dev *virt_dev)
{
	uint64_t dev_id_num;

	spin_lock_init(&virt_dev->flags_lock);

	virt_dev->rd_only = DEF_RD_ONLY;
	virt_dev->removable = DEF_REMOVABLE;
	virt_dev->thin_provisioned = DEF_THIN_PROVISIONED;

	virt_dev->block_size = DEF_DISK_BLOCKSIZE;
	virt_dev->block_shift = DEF_DISK_BLOCKSIZE_SHIFT;
	virt_dev->media_exists = DEF_MEDIA_EXISTS;

	virt_dev->wt_flag = DEF_WRITE_THROUGH;
	virt_dev->nv_cache = DEF_NV_CACHE;
	virt_dev->o_direct_flag = DEF_O_DIRECT;

	dev_id_num = gen_dev_id_num(virt_dev->name);

	snprintf(virt_dev->t10_dev_id, sizeof(virt_dev->t10_dev_id),
		"%llx-%s", dev_id_num, virt_dev->name);
	TRACE_DBG("t10_dev_id %s", virt_dev->t10_dev_id);

	scnprintf(virt_dev->usn, sizeof(virt_dev->usn), "%llx", dev_id_num);
	TRACE_DBG("usn %s", virt_dev->usn);

	virt_dev->command_set_version = 0x04C0; /* SBC-3 */

	return 0;
}

void pcimem_del_virtual_device(struct scst_pcimem_dev *virt_dev)
{
	return;
}

int pcimem_init_virtual_dev(void)
{
	p_dh_trace_flag = get_trace_flag_ptr();
	dh_trace_flag = trace_flag; // unused

	return 0;
}

void pcimem_exit_virtual_dev(void)
{
	return;
}

/* end of scst_pcimem_dummy.c */
