/*
 *  scst_pcimem_ipc.c
 *
 *  Copyright (C) 2004 - 2011 Vladislav Bolkhovitin <vst@vlnb.net>
 *  Copyright (C) 2004 - 2005 Leonid Stoljar
 *  Copyright (C) 2007 Ming Zhang <blackmagic02881 at gmail dot com>
 *  Copyright (C) 2007 Ross Walker <rswwalker at hotmail dot com>
 *  Copyright (C) 2007 - 2010 ID7 Ltd.
 *  Copyright (C) 2010 - 2011 SCST Ltd.
 *
 *  SCSI disk (type 0) dev handler using PCI-memory space (PCIMEM).
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, version 2
 *  of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>

#include <asm/unaligned.h>

#ifdef INSIDE_KERNEL_TREE
#include <scst/scst.h>
#else
#include "scst.h"
#endif

#include "../../../../sony/ipcif_scstlib/ipcif_scstlib.h"

#include "scst_priv.h"
#include "scst_dev_handler.h"
#include "scst_mem.h"
#include "scst_pcimem.h"
#include "scst_ipc.h"

#if defined(CONFIG_SCST_DEBUG) || defined(CONFIG_SCST_TRACING)

static unsigned long *p_dh_trace_flag;

#undef trace_flag
#define trace_flag (*p_dh_trace_flag)

#endif /* defined(CONFIG_SCST_DEBUG) || defined(CONFIG_SCST_TRACING) */


static struct sgv_pool_obj *pcimem_create_sg_obj(gfp_t gfp_mask, struct scst_cmd *cmd);
static void pcimem_proc_cmdresp(void *priv, unsigned int tag, unsigned int resp_code, unsigned int sc, unsigned int asc, unsigned int ascq, unsigned int add_inf, unsigned int add_spec);
static void pcimem_proc_dmardy(void *priv, unsigned int tag, struct ipcscst_areainfo *area, int num_area, int direction);
static void pcimem_proc_reset_resp(void *priv);
static int pcimem_send_cmdreq(struct scst_cmd *cmd);
static int pcimem_send_dmadone(struct scst_cmd *cmd);

static struct ipcscst_funcs ipcscst_funcs = {
	pcimem_proc_cmdresp,
	pcimem_proc_dmardy,
	pcimem_proc_reset_resp,
};


static void pcimem_set_sense_info(struct scst_cmd *cmd, unsigned int resp_code, unsigned int sc, unsigned int asc, unsigned int ascq, unsigned int add_inf, unsigned int add_spec)
{
	u8 *p;

	TRACE_ENTRY();

	scst_set_cmd_error(cmd, sc, asc, ascq);

	TRACE(TRACE_DEBUG, "cmd->sense_valid_len=0x%x", cmd->sense_valid_len);

	if (resp_code & 0x80) {
		cmd->sense[0] |= 0x80;
	}

	if (cmd->sense[0] != resp_code) {
		TRACE(TRACE_DEBUG, "cmd->sense[0](0x%02x)<=resp_code(0x%02x)", cmd->sense[0], resp_code);
		cmd->sense[0] = resp_code;
	}

	if (cmd->sense_valid_len >= 6) {
		p = (u8 *)&add_inf;
		cmd->sense[3] = p[0];
		cmd->sense[4] = p[1];
		cmd->sense[5] = p[2];
		cmd->sense[6] = p[3];
	}
	if (cmd->sense_valid_len >= 17) {
		p = (u8 *)&add_spec;
		cmd->sense[15] = p[0];
		cmd->sense[16] = p[1];
		cmd->sense[17] = p[2];
	}

	TRACE(TRACE_DEBUG, "resp_code=0x%02x,info=0x%02x%02x%02x%02x,spec=0x%02x%02x%02x",
		  cmd->sense[0],
		  cmd->sense[3], cmd->sense[4], cmd->sense[5], cmd->sense[6],
		  cmd->sense[15], cmd->sense[16], cmd->sense[17]);

	TRACE_EXIT();
	return;
}

static void pcimem_exec_request_sense(struct scst_cmd *cmd)
{
	int32_t length, sl;
	uint8_t *address;
	uint8_t b[SCST_STANDARD_SENSE_LEN];

	TRACE_ENTRY();

	sl = scst_set_sense(b, sizeof(b), cmd->dev->d_sense,
		SCST_LOAD_SENSE(scst_sense_no_sense));

	length = scst_get_buf_full(cmd, &address);
	TRACE_DBG("length %d", length);
	if (length < 0) {
		PRINT_ERROR("scst_get_buf_full() failed: %d)", length);
		scst_set_cmd_error(cmd,
			SCST_LOAD_SENSE(scst_sense_hardw_error));
		goto out;
	}

	length = min(sl, length);
	memcpy(address, b, length);
	scst_set_resp_data_len(cmd, length);

	scst_put_buf_full(cmd, address);

out:
	TRACE_EXIT();
	return;
}

int pcimem_parse(struct scst_cmd *cmd)
{
	struct scst_device *dev = cmd->dev;
	struct scst_pcimem_dev *virt_dev = dev->dh_priv;
	int res = SCST_CMD_STATE_DEFAULT;
	int is_vendor_unique_cmd = (cmd->op_flags & SCST_INFO_VALID) == 0;

	TRACE_ENTRY();

	cmd->status = 0;

	memcpy(cmd->cdb_org, cmd->cdb, sizeof(cmd->cdb_org));

	if (is_vendor_unique_cmd) {
		cmd->data_direction = cmd->expected_data_direction;
		cmd->bufflen = cmd->expected_transfer_len;
		cmd->timeout = 60 * 60 * HZ;
		cmd->op_name = "VENDOR UNIQUE";
		cmd->cdb_len = 16;
		cmd->op_flags |= SCST_INFO_VALID;

		TRACE_DBG("opcode %s[0x%02x] direct %d flags %d transfer_len %d",
				  cmd->op_name, cmd->cdb[0], cmd->data_direction, cmd->op_flags, cmd->bufflen);
	} else {
		switch (dev->type) {
		case TYPE_MOD:
			scst_modisk_generic_parse(cmd, pcimem_get_block_shift);
			break;
		case TYPE_MEDIUM_CHANGER:
			scst_changer_generic_parse(cmd, NULL);
			break;
		case TYPE_DISK:
		default:
			scst_sbc_generic_parse(cmd, pcimem_get_block_shift);
			break;
		}

	}

	if (scst_is_serialized_cmd(cmd) &&
		(cmd->data_direction == SCST_DATA_WRITE)) {
		/*
		 *  serialized_cmd block the other command(s).
		 *  Don't defer in scst_send_for_exec(),
		 *  even if the blocked command is in QUEUE_HEAD_OF_QUEUE.
		 */
		cmd->not_to_be_deferred = 1;
	}

	if (unlikely(((cmd->cdb[0] == 0x16) || /* RESERVE */
				  (cmd->cdb[0] == 0x17) || /* RELEASE */
				  (cmd->cdb[0] == 0x56) || /* RESERVE(10) */
				  (cmd->cdb[0] == 0x57) || /* RELEASE(10) */
				  (cmd->cdb[0] == 0x5F)) /* PERSISTENT RESERVE OUT */
				 )) {

		PRINT_INFO("reserve/release opcode=0x%02x received (lun=%llu, initiator %s)", cmd->cdb[0], cmd->lun, cmd->sess->initiator_name);

	}

	/* reservation inhibit, if unsupported */
	if (unlikely(((cmd->cdb[0] == 0x16) || /* RESERVE */
				  (cmd->cdb[0] == 0x17) || /* RELEASE */
				  (cmd->cdb[0] == 0x56) || /* RESERVE(10) */
				  (cmd->cdb[0] == 0x57) || /* RELEASE(10) */
				  (cmd->cdb[0] == 0x5E) || /* PERSISTENT RESERVE IN */
				  (cmd->cdb[0] == 0x5F)) /* PERSISTENT RESERVE OUT */
				 && virt_dev->no_reserve)) {

		PRINT_ERROR("opcode=0x%02x is unsupported (lun=%llu, initiator %s)", cmd->cdb[0], cmd->lun, cmd->sess->initiator_name);
		scst_set_cmd_error(cmd,
						   SCST_LOAD_SENSE(scst_sense_invalid_opcode));
		scst_set_cmd_abnormal_done_state(cmd);
	}

	TRACE_EXIT_RES(res);
	return res;
}

static void pcimem_exec_datain(struct scst_cmd *cmd)
{
	struct scst_device *dev = cmd->dev;
	struct scst_pcimem_dev *virt_dev = dev->dh_priv;
	int ret;

	TRACE_ENTRY();

	TRACE(TRACE_DEBUG, "bufflen=0x%x", cmd->bufflen);

	if (unlikely(virt_dev->obmbxid == -1)) {
		PRINT_ERROR("%s", "invalid obmbxid.\n");
		goto out;
	}

	if (unlikely(SCST_SENSE_VALID(cmd->sense))) {
		PRINT_ERROR("%s", "sense-code has been set.\n");
		goto out;
	}

	ret = pcimem_send_cmdreq(cmd);
	if (unlikely(ret < 0)) {
		//		PRINT_ERROR("pcimem_send_cmdreq() error(%d).\n", ret);
		goto out;
	}

	do {
		ret = ipcscst_recv(virt_dev->obmbxid, cmd, &ipcscst_funcs);
		if (unlikely(ret < 0)) {
			PRINT_ERROR("ipcscst_recv() error(%d).\n", ret);
			goto out;
		}
	} while (virt_dev->ipc_active_cmd == cmd);

out:
	TRACE_EXIT();
	return;
}

static void pcimem_exec_dataout(struct scst_cmd *cmd)
{
	struct scst_device *dev = cmd->dev;
	struct scst_pcimem_dev *virt_dev = dev->dh_priv;
	int ret;

	TRACE_ENTRY();

	TRACE(TRACE_DEBUG, "bufflen=0x%x", cmd->bufflen);

	if (unlikely(virt_dev->obmbxid == -1)) {
		PRINT_ERROR("%s", "invalid obmbxid.\n");
		goto out;
	}

	if (unlikely(SCST_SENSE_VALID(cmd->sense))) {
		PRINT_ERROR("%s", "sense-code has been set.\n");
		goto out;
	}

	ret = pcimem_send_dmadone(cmd);
	if (ret < 0) {
		PRINT_ERROR("pcimem_send_dmadone() error(%d)\n", ret);
		goto out;
	}

	ret = ipcscst_recv(virt_dev->obmbxid, cmd, &ipcscst_funcs);
	if (unlikely(ret < 0)) {
		PRINT_ERROR("ipcscst_recv() error(%d).\n", ret);
		goto out;
	}

out:
	TRACE_EXIT();
	return;
}

static void pcimem_exec_nodata(struct scst_cmd *cmd)
{
	struct scst_device *dev = cmd->dev;
	struct scst_pcimem_dev *virt_dev = dev->dh_priv;
	int ret;

	TRACE_ENTRY();

	TRACE(TRACE_DEBUG, "bufflen=0x%x", cmd->bufflen);

	if (virt_dev->exmount &&
		(cmd->cdb[0] == TEST_UNIT_READY) &&
		!virt_dev->no_reserve && // temporary
		!cmd->has_reservation) {

		if (cmd->tgt_dev == virt_dev->reserve_owner) {
			virt_dev->reserve_owner = NULL;
		}

		goto reserved_by_other;
	}

	if (unlikely(virt_dev->obmbxid == -1)) {
		PRINT_ERROR("%s", "invalid obmbxid.\n");
		goto out;
	}

	ret = pcimem_send_cmdreq(cmd);
	if (unlikely(ret < 0)) {
		//		PRINT_ERROR("pcimem_send_cmdreq() error(%d).\n", ret);
		goto out;
	}

	do {
		ret = ipcscst_recv(virt_dev->obmbxid, cmd, &ipcscst_funcs);
		if (unlikely(ret < 0)) {
			PRINT_ERROR("ipcscst_recv() error(%d).\n", ret);
			goto out;
		}
	} while (virt_dev->ipc_active_cmd == cmd);

	if (virt_dev->exmount &&
		(cmd->cdb[0] == TEST_UNIT_READY) &&
		(cmd->tgt_dev != virt_dev->reserve_owner)) {

		virt_dev->reserve_owner = cmd->tgt_dev;

		if (!cmd->sense) {
			scst_set_cmd_error(cmd, UNIT_ATTENTION, 0x28, 0x00);
		}
	}

out:
	TRACE_EXIT();
	return;

reserved_by_other:
	TRACE(TRACE_DEBUG, "TEST_UNIT_READY from non-reserved Initiator(%s), return MediaNotPresent\n", cmd->sess->initiator_name);
	scst_set_cmd_error(cmd, NOT_READY, 0x3a, 0x00);

	goto out;
}

static void pcimem_exec_local_cmd(struct scst_cmd *cmd)
{
	switch (cmd->cdb[0]) {
	case REQUEST_SENSE:
		pcimem_exec_request_sense(cmd);
		break;
	case REPORT_LUNS:
		break;
	default:
		break;
	}

	return;
}

/*
 *  reserve ipcif-MBX
 *  return 0 = MBX is already used by other session
 *         1 = MBX is idle, or used by same session. Be sure to check if ipc_active_cmd is updated.
 *
 *  don't call under atomic context.
 */
static int pcimem_check_reserve_mbx(struct scst_pcimem_dev *virt_dev, struct scst_cmd *cmd)
{
	int res = 0;

	spin_lock(&virt_dev->active_cmdinfo_lock);

	if (virt_dev->ipc_active_cmd) {
		if (virt_dev->ipc_active_cmd->tgt_dev == cmd->tgt_dev) {
			/* same session, return without updating ipc_active_cmd */
			res = 1;
		}

		goto out;
	}

	virt_dev->ipc_active_cmd = cmd;
	virt_dev->last_updated = jiffies;
	cmd->dh_sn = virt_dev->curr_sn++;

	res = 1;

out:
	spin_unlock(&virt_dev->active_cmdinfo_lock);

	return res;
}

/*
 *  release ipcif-MBX
 * 
 * don't call under atomic context.
 */
static void pcimem_release_mbx(struct scst_pcimem_dev *virt_dev, struct scst_cmd *cmd)
{
	spin_lock(&virt_dev->active_cmdinfo_lock);

	if ((cmd == NULL) ||
		(cmd == virt_dev->ipc_active_cmd)) {
		virt_dev->ipc_active_cmd = NULL;
	}

	spin_unlock(&virt_dev->active_cmdinfo_lock);

	wake_up_interruptible(&virt_dev->mbx_waitQ);

	return;
}

int pcimem_do_job(struct scst_cmd *cmd)
{
	int rc, res;
	int opcode = cmd->cdb[0];

	TRACE_ENTRY();

	TRACE(TRACE_DEBUG, "opcode=0x%02x, retry=%d", opcode, cmd->dh_exec_retry);

	switch (cmd->queue_type) {
	case SCST_CMD_QUEUE_ORDERED:
		TRACE(TRACE_ORDER, "ORDERED cmd %p (op %x)", cmd, cmd->cdb[0]);
		break;
	case SCST_CMD_QUEUE_HEAD_OF_QUEUE:
		TRACE(TRACE_ORDER, "HQ cmd %p (op %x)", cmd, cmd->cdb[0]);
		break;
	default:
		break;
	}

	if (scst_is_cmd_local(cmd)) {
		pcimem_exec_local_cmd(cmd);
		goto out_compl;
	}

	rc = scst_check_local_events(cmd);
	if (unlikely(rc != 0)) {
		goto out_done;
	}

	cmd->status = 0;
	cmd->msg_status = 0;
	cmd->host_status = DID_OK;
	cmd->driver_status = 0;

	if (cmd->data_direction == SCST_DATA_WRITE) {
		pcimem_exec_dataout(cmd);
	} else if (cmd->data_direction == SCST_DATA_READ) {
		pcimem_exec_datain(cmd);
	} else if (cmd->data_direction == SCST_DATA_BIDI) {
		PRINT_ERROR("%s", "can't exec bidirection transfer.\n");
		scst_set_cmd_error(cmd,
						   SCST_LOAD_SENSE(scst_sense_hardw_error));
		goto out_compl;
	} else {
		pcimem_exec_nodata(cmd);
	}

out_compl:
	cmd->completed = 1;

out_done:
	cmd->scst_cmd_done(cmd, SCST_CMD_STATE_DEFAULT, SCST_CONTEXT_SAME);

	res = SCST_EXEC_COMPLETED;

	cmd->dh_exec_retry++;

	TRACE_EXIT_RES(res);
	return res;
}


int scst_pcimem_allocator_init(struct scst_pcimem_dev *virt_dev)
{
	int res = 0;

	return res;
}

static void pcimem_proc_cmdresp(void *priv, unsigned int tag, unsigned int resp_code, unsigned int sc, unsigned int asc, unsigned int ascq, unsigned int add_inf, unsigned int add_spec)
{
	struct scst_cmd *cmd = priv;
	struct scst_device *dev = cmd->dev;
	struct scst_pcimem_dev *virt_dev = dev->dh_priv;

	TRACE_ENTRY();

	TRACE(TRACE_SCSI, "cmd=%p,rc=0x%02x,sc=0x%02x,asc=0x%02x,ascq=0x%02x", cmd, resp_code, sc, asc, ascq);

	if (sc) {
		pcimem_set_sense_info(cmd, resp_code, sc, asc, ascq, add_inf, add_spec);
	} else if (unlikely((cmd->dh_sg_cnt == 0) && (cmd->data_direction != SCST_DATA_NONE))) {
		scst_set_resp_data_len(cmd, 0);
	}

	scstipc_check_save_cache(cmd);

	TRACE(TRACE_DEBUG, "pcimem_release_mbx(cmd=%p, ibmbxid=%d)", cmd, virt_dev->ibmbxid);

	pcimem_release_mbx(virt_dev, cmd);

	TRACE_EXIT();

	return;
}

static void pcimem_proc_dmardy(void *priv, unsigned int tag, struct ipcscst_areainfo *area, int num_area, int direction)
{
	struct scst_cmd *cmd = priv;
	int size = 0;
	int n;
	int i, j;

	TRACE_ENTRY();

	if ((num_area <= 0) || (num_area > MAX_SGV_CNT - cmd->sg_cnt)) {
		PRINT_ERROR("invalid num_area(=%d)\n", num_area);
		cmd->sg_cnt = 0;
		goto out;
	}

	if ((cmd->data_direction == SCST_DATA_NONE) ||
		(!cmd->sgv) ||
		(!cmd->sg)) {
		PRINT_ERROR("can't process dmardy for cmd %p (opcode=0x%02x, direction=%d, bufflen=%d)\n", cmd, cmd->cdb[0], cmd->data_direction, cmd->bufflen);
		scst_set_cmd_error(cmd,
						   SCST_LOAD_SENSE(scst_sense_hardw_error));
		goto out;
	}

	if (cmd->sg_cnt) {
		size = cmd->resp_data_len;
	}

	TRACE_MEM("cmd=%p", cmd);

	for (i = 0; i < num_area; i++) {
		cmd->dh_sg[cmd->dh_sg_cnt].dma_address = area[i].addr;
		cmd->dh_sg[cmd->dh_sg_cnt].dma_length = area[i].size;

		TRACE_MEM("dh_sg[%d],offset=0x%08lx,len=0x%08lx", cmd->dh_sg_cnt, area[i].addr, area[i].size);

		n = pcimem_add_sg_entries(&cmd->sg[cmd->sg_cnt], MAX_SGV_CNT - cmd->sg_cnt, area[i].addr, area[i].size);
		if (n <= 0) {
			PRINT_ERROR("pcimem_add_sg_entries(offset=0x%08lx,size=0x%08lx) error", area[i].addr, area[i].size);
			goto out;
		}

		for (j = 0; j < n; j++) {
			TRACE(TRACE_DEBUG, "%d,dma_addr=0x%08x,len=0x%08x", cmd->sg_cnt + j, cmd->sg[cmd->sg_cnt + j].dma_address, cmd->sg[cmd->sg_cnt + j].dma_length);
		}

		size += area[i].size;
		cmd->sg_cnt += n;

		cmd->dh_sg_cnt++;
	}

	if (size < cmd->bufflen) {
		TRACE(TRACE_DEBUG, "size short(reqsize=0x%x,alloced=0x%x)", cmd->bufflen, size);

		if (cmd->data_direction & SCST_DATA_READ) {
			scst_set_resp_data_len(cmd, size);
		}
	}

	cmd->no_local_mem = 1;

out:
	TRACE_EXIT();
	return;
}

static void pcimem_proc_reset_resp(void *priv)
{
	struct scst_cmd *cmd = priv;
	struct scst_device *dev = cmd->dev;
	struct scst_pcimem_dev *virt_dev = dev->dh_priv;

	TRACE_ENTRY();

	pcimem_release_mbx(virt_dev, NULL);

	TRACE_EXIT();
	return;
}

static int pcimem_send_cmdreq(struct scst_cmd *cmd)
{
	struct scst_device *dev = cmd->dev;
	struct scst_pcimem_dev *virt_dev = dev->dh_priv;
	int res = 0;
	int ret;

	TRACE_ENTRY();

	TRACE(TRACE_SCSI, "cmd=%p", cmd);
	PRINT_BUFF_FLAG(TRACE_SCSI, "Requesting CDB",
		cmd->cdb_org, cmd->cdb_len);

	if (unlikely(virt_dev->ibmbxid == -1)) {
		PRINT_ERROR("%s", "invalid ibmbxid.\n");
		return -1;
	}

	TRACE(TRACE_DEBUG, "wait_event_interruptible_timeout(cmd=%p, ibmbxid=%d, opcode=0x%02x, initiator=%s)", cmd, virt_dev->ibmbxid, cmd->cdb[0], cmd->sess->initiator_name);

	ret = wait_event_interruptible_timeout(virt_dev->mbx_waitQ, pcimem_check_reserve_mbx(virt_dev, cmd), cmd->timeout);

	TRACE(TRACE_DEBUG, "wait_event_interruptible_timeout(cmd=%p)=%d", cmd, ret);

	if ((ret == 0) || /* timeout */
		(ret == -ERESTARTSYS) || /* interrupted */
		(virt_dev->ipc_active_cmd != cmd)) { /* couldn't reserve MBX */
		if (scstipc_check_read_cache(cmd) == 0) {
			TRACE(TRACE_DEBUG, "cmd %p cache found", cmd);

		} else {
			PRINT_INFO("cmd %p (op %02x, LUN %llu initiator %s) cache not found, returning BUSY status", cmd, cmd->cdb[0], cmd->lun, cmd->sess->initiator_name);
			//			TRACE(TRACE_DEBUG, "cmd %p cache not found, returning BUSY status", cmd);
			scst_set_busy(cmd);
			scst_set_cmd_abnormal_done_state(cmd);
		}

		res = -1;
		goto out;
	}

	ipcscst_flush_recvmsg(virt_dev->obmbxid);

	ret = ipcscst_send_cmdreq(virt_dev->ibmbxid, cmd->dh_sn, cmd->cdb_org);
	if (unlikely(ret < 0)) {
		PRINT_ERROR("ipcscst_send_cmdreq() error(%d).\n", ret);
		goto out_fail;
	}

out:
	TRACE_EXIT_RES(res);
	return res;

out_fail:
	pcimem_release_mbx(virt_dev, cmd);
	res = -1;
	goto out;
}

static int pcimem_send_dmadone(struct scst_cmd *cmd)
{
	struct scst_device *dev = cmd->dev;
	struct scst_pcimem_dev *virt_dev = dev->dh_priv;
	struct ipcscst_areainfo areainfo[MAX_SGV_CNT];
	int direction;
	int ret;
	int i;

	switch (cmd->data_direction) {
	case SCST_DATA_READ:
		direction = IPCSCST_DATA_READ;
		break;
	case SCST_DATA_WRITE:
		direction = IPCSCST_DATA_WRITE;
		break;
	default:
		direction = 0;
		break;
	}

	TRACE_MEM("cmd=%p,cmd->dh_sg_cnt=%d", cmd, cmd->dh_sg_cnt);

	for (i = 0; i < cmd->dh_sg_cnt; i++) {
		areainfo[i].addr = cmd->dh_sg[i].dma_address;
		areainfo[i].size = cmd->dh_sg[i].dma_length;
	}

	ret = ipcscst_send_dmadone(virt_dev->ibmbxid, cmd->dh_sn, areainfo, cmd->dh_sg_cnt, direction);
	if (unlikely(ret < 0)) {
		PRINT_ERROR("ipcscst_send_dmadone() error(%d).", ret);
		return ret;
	}

	return 0;
}

static struct sgv_pool_obj *pcimem_create_sg_obj(gfp_t gfp_mask, struct scst_cmd *cmd)
{
	struct sgv_pool_obj *obj;
	int sz;

	sz = sizeof(*obj) + MAX_SGV_CNT * sizeof(obj->sg_entries[0]);

	obj = kmalloc(sz, gfp_mask);
	if (unlikely(obj == NULL)) {
		TRACE(TRACE_OUT_OF_MEM, "Allocation of "
			  "sgv_pool_obj failed (size %d)", sz);
		return NULL;
	}

	memset(obj, 0, sizeof(*obj));

	obj->owner_pool = NULL;
	obj->cache_num = -1;
	obj->allocator_priv = cmd;

	obj->sg_entries = obj->sg_entries_data;
	sg_init_table(obj->sg_entries, MAX_SGV_CNT);

	return obj;
}

static void pcimem_free_sg_entries(struct sgv_pool_obj *obj)
{
	kfree(obj);

	return;
}

static int pcimem_create_sg_objs(struct scst_cmd *cmd)
{
	struct scst_tgt_dev *tgt_dev = cmd->tgt_dev;
	struct sgv_pool_obj *obj;
	gfp_t gfp_mask;
	int atomic = scst_cmd_atomic(cmd);
	int res = -ENOMEM;

	TRACE_ENTRY();

	gfp_mask = tgt_dev->gfp_mask | (atomic ? GFP_ATOMIC : GFP_KERNEL);

	obj = pcimem_create_sg_obj(gfp_mask, cmd);
	if (!obj) {
		goto out;
	}

	cmd->dh_sg_cnt = 0;
	cmd->dh_sgv = obj;
	cmd->dh_sg = obj->sg_entries;

	obj = pcimem_create_sg_obj(gfp_mask, cmd);
	if (!obj) {
		goto out_dh_sg_free;
	}

	cmd->sg_cnt = 0;
	cmd->sgv = obj;
	cmd->sg = obj->sg_entries;

	if (unlikely(cmd->sg_cnt > tgt_dev->max_sg_cnt))
		goto out_sg_free;

	if (cmd->data_direction != SCST_DATA_BIDI)
		goto success;

	obj = pcimem_create_sg_obj(gfp_mask, cmd);
	if (!obj) {
		goto out_sg_free;
	}

	cmd->out_sg_cnt = 0;
	cmd->out_sgv = obj;
	cmd->out_sg = obj->sg_entries;

	if (unlikely(cmd->out_sg_cnt > tgt_dev->max_sg_cnt))
		goto out_out_sg_free;

success:
	cmd->dh_data_buf_alloced = 1;
	res = 0;

out:
	TRACE_EXIT_RES(res);
	return res;

out_out_sg_free:
	pcimem_free_sg_entries(cmd->out_sgv);
	cmd->out_sgv = NULL;
	cmd->out_sg = NULL;
	cmd->out_sg_cnt = 0;

out_sg_free:
	pcimem_free_sg_entries(cmd->sgv);
	cmd->sgv = NULL;
	cmd->sg = NULL;
	cmd->sg_cnt = 0;

out_dh_sg_free:
	pcimem_free_sg_entries(cmd->dh_sgv);
	cmd->dh_sgv = NULL;
	cmd->dh_sg = NULL;
	cmd->dh_sg_cnt = 0;

	goto out;
}

int pcimem_alloc_data_buf(struct scst_cmd *cmd)
{
	int res = SCST_CMD_STATE_DEFAULT;

	TRACE_ENTRY();

	TRACE_MEM("cmd=%p,opcode=0x%02x,queue=%d", cmd, cmd->cdb[0], cmd->queue_type);

	if (scst_is_cmd_local(cmd)) {
		goto out;
	}

	if (cmd->data_direction & SCST_DATA_WRITE) {
		/*
		 *  if device is blocked, try again...
		 *  (cmd will be made waiting in 'blocked_cmd_list' until device unblocked)
		 */
		if (unlikely(scst_check_blocked_dev_on_alloc(cmd))) {
			res = SCST_CMD_STATE_PREPARE_SPACE;
			goto out;
		}
	}

	pcimem_create_sg_objs(cmd);

	if (cmd->data_direction & SCST_DATA_WRITE) {
		struct scst_device *dev = cmd->dev;
		struct scst_pcimem_dev *virt_dev = dev->dh_priv;
		int ret;

		TRACE(TRACE_DEBUG, "calling scst_pre_check_local_events(opcode=0x%02x)", cmd->cdb[0]);

		ret = scst_pre_check_local_events(cmd);
		if (unlikely(ret != 0)) {
			res = SCST_CMD_STATE_PRE_DEV_DONE;
			goto out;
		}

		ret = pcimem_send_cmdreq(cmd);
		if (unlikely(ret < 0)) {
			//			PRINT_ERROR("pcimem_send_cmdreq() error(%d).\n", ret);
			res = SCST_CMD_STATE_PRE_DEV_DONE;
			goto out;
		}

		ret = ipcscst_recv(virt_dev->obmbxid, cmd, &ipcscst_funcs);
		if (unlikely(ret < 0)) {
			PRINT_ERROR("ipcscst_recv() error(%d).\n", ret);
			res = SCST_CMD_STATE_STOP;
			goto out;
		}

		if (cmd->status) {
			/*
			 *  if allocation failed, skip transfer...
			 */
			res = SCST_CMD_STATE_PRE_DEV_DONE;
		}
	}

out:
	TRACE_EXIT_RES(res);
	return res;
}

void pcimem_on_free_cmd(struct scst_cmd *cmd)
{
	int ret;

	TRACE_ENTRY();

	TRACE_MEM("cmd=%p,opcode=0x%02x", cmd, cmd->cdb[0]);

	if (!cmd->dh_data_buf_alloced) {
		goto out;
	}

	if (cmd->out_sgv != NULL) {
		pcimem_free_sg_entries(cmd->out_sgv);
		cmd->out_sgv = NULL;
		cmd->out_sg_cnt = 0;
		cmd->out_sg = NULL;
		cmd->out_bufflen = 0;
	}

	if ((cmd->sg_cnt) &&
		(cmd->data_direction & SCST_DATA_READ)) {
		if (cmd->no_local_mem) {
			if (in_atomic()) {
				ret = scstipc_add_dmadone_work(cmd);
				if (ret < 0) {
					PRINT_ERROR("pcimem_add_dmadone_event() error(%d)\n", ret);
				}
			} else {
				ret = pcimem_send_dmadone(cmd);
				if (ret < 0) {
					PRINT_ERROR("pcimem_send_dmadone() error(%d)\n", ret);
				}
			}
		} else {
			scstipc_cache_free_local_buf(cmd->sg, cmd->sg_cnt);
		}
	}

	pcimem_free_sg_entries(cmd->sgv);
	cmd->sgv = NULL;
	cmd->sg_cnt = 0;
	cmd->sg = NULL;
	cmd->bufflen = 0;
	cmd->data_len = 0;

	pcimem_free_sg_entries(cmd->dh_sgv);
	cmd->dh_sgv = NULL;
	cmd->dh_sg_cnt = 0;
	cmd->dh_sg = NULL;

out:
	TRACE_EXIT();
	return;
}

int pcimem_task_mgmt_fn(struct scst_mgmt_cmd *mcmd,
	struct scst_tgt_dev *tgt_dev)
{
	TRACE_ENTRY();

	if ((mcmd->fn == SCST_LUN_RESET) || (mcmd->fn == SCST_TARGET_RESET)) {
		/* Restore default values */
		struct scst_device *dev = tgt_dev->dev;
		struct scst_pcimem_dev *virt_dev = dev->dh_priv;

		dev->tst = DEF_TST;
		dev->d_sense = DEF_DSENSE;
		if (virt_dev->wt_flag && !virt_dev->nv_cache)
			dev->queue_alg = DEF_QUEUE_ALG_WT;
		else
			dev->queue_alg = DEF_QUEUE_ALG;
		dev->swp = DEF_SWP;
		dev->tas = DEF_TAS;

		//		ipcscst_send_reset(virt_dev->ibmbxid);
	} else if (mcmd->fn == SCST_PR_ABORT_ALL) {
		//		struct scst_device *dev = tgt_dev->dev;
		//		struct scst_pcimem_dev *virt_dev = dev->dh_priv;

		//		ipcscst_send_reset(virt_dev->ibmbxid);
	}

	TRACE_EXIT();
	return SCST_DEV_TM_NOT_COMPLETED;
}

int pcimem_add_virtual_device(struct scst_pcimem_dev *virt_dev)
{
	scstipc_init_dev_cache(virt_dev);
	virt_dev->ibmbxid = -1;
	virt_dev->obmbxid = -1;
	virt_dev->curr_sn = 0;
	virt_dev->ipc_active_cmd = NULL;
	spin_lock_init(&virt_dev->active_cmdinfo_lock);
	init_waitqueue_head(&virt_dev->mbx_waitQ);

	return 0;
}

void pcimem_del_virtual_device(struct scst_pcimem_dev *virt_dev)
{
	scstipc_exit_dev_cache(virt_dev);

	return;
}

int pcimem_init_virtual_dev(void)
{
	p_dh_trace_flag = get_trace_flag_ptr();
	dh_trace_flag = trace_flag; // unused

	scstipc_init_delayed_work();
	scstipc_init_cache_ctrl();

	return 0;
}

void pcimem_exit_virtual_dev(void)
{
	scstipc_exit_cache_ctrl();
	scstipc_exit_delayed_work();

	return;
}

/* end of scst_pcimem_ipc.c */
